/* gl_md3.c 
 * Based on code from the Aftershock 3D rendering engine 
 * Copyright (C) 1999 Stephen C. Taylor 
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
 * GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 */ 
#include "quakedef.h" 
#ifdef Q3MODELS 
#include "gl_md3.h" 

#define TEMPHACK 1 
#ifdef TEMPHACK 
extern cvar_t temp1; 
#endif 

void R_MD3TagRotate (entity_t *e, model_t *tagmodel, char *tagname) 
{ 
   int i; 
   md3tag_t *tag = NULL; 
   md3header_t *model = Mod_Extradata(tagmodel); 
   float m[16]; 

   for (i=0; i<model->num_tags; i++) 
   { 
      md3tag_t *tags = (md3tag_t *)((byte *)model + model->tag_offs); 
      if(Q_strcmp(tags[i].name, tagname)==0) 
#if TEMPHACK 
         if(temp1.value > model->num_frames) 
            tag = &tags[(model->num_frames-1) * model->num_tags + i]; 
         else 
            tag = &tags[(int)temp1.value * model->num_tags + i]; 
#else 
         if(e->frame > model->num_frames) 
            tag = &tags[(model->num_frames-1) * model->num_tags + i]; 
         else 
            tag = &tags[e->frame * model->num_tags + i]; 
#endif 
   } 

   if(!tag) 
   { 
      Con_Printf("Tag not found in %s : %s\n", tagmodel->name, tagname); 
      return; 
   } 

   m[0] = tag->rot[0][0];   m[4] = tag->rot[1][0];   m[8] = tag->rot[2][0];   m[12] = tag->pos[0]; 
   m[1] = tag->rot[0][1];   m[5] = tag->rot[1][1];   m[9] = tag->rot[2][1];   m[13] = tag->pos[1]; 
   m[2] = tag->rot[0][2];   m[6] = tag->rot[1][2];   m[10]= tag->rot[2][2];   m[14] = tag->pos[2]; 
   m[3] = 0;            m[7] = 0;            m[11]= 0;            m[15] = 1; 

   glMultMatrixf(m); 
} 

void R_DrawQ3Model(entity_t *e) 
{ 
   md3header_mem_t *model = Mod_Extradata (e->model); 
   int i; 
   int j; 
   //int *tris; 
   md3surface_mem_t *surf; 
   extern vec3_t lightcolor; 


   if (*(long *)model->id != MD3IDHEADER){ 
      Con_Printf("MD3 bad model for: %s\n", model->filename); 
      return; 
   } 

//#ifdef EXTENDQC 
//   glColor4f (l, l, l, currententity->alpha); 
//#else 
   glColor3fv(lightcolor); 

   glPushMatrix(); 
   R_RotateForEntity(e); 
    
   surf = model->surfaces; 
   for (i = 0; i < model->num_surfaces; i++, surf = surf->nextsurf) 
   { 
      if (*(long *)surf->id != MD3IDHEADER){ 
         Con_Printf("MD3 bad surface for: %s\n", model->filename); 
         continue; 
      } 
      //int frameoffset;//numframes, frameoffset; 
/*#if TEMPHACK 
      frameoffset = temp1.value; 
#else 
      frameoffset = e->frame; 
#endif*/ 

      if (surf->num_surf_shaders!=0) 
         GL_Bind(surf->shaders[(e->skinnum%surf->num_surf_shaders)].texnum); 
      else 
         GL_Bind(0); 
      glBegin(GL_TRIANGLES); 
         //tris = surf->tris; 
         //numpoints = surf->num_surf_tris*3; 
          
         //if(e->frame > model->num_frames) 
         //   frameoffset = model->num_frames-1; 

         //frameoffset *= surf->num_surf_verts; 

         //for each triangle 
         for (j = 0; j < surf->num_surf_tris; j++) 
         { 
            //draw the poly 
            int k; 
            for (k=0; k < 3; k++){ 
               glTexCoord2f(surf->tc[surf->tris[j].tri[k]].s, surf->tc[surf->tris[j].tri[k]].t);//((vec2_t *)((byte *)surf + surf->tc_offs))[*tris]); 
               glVertex3fv(surf->verts[surf->tris[j].tri[k]].vec);//((vec3_t *)((byte *)model + surf->vert_offs))[*tris + frameoffset]); 
            } 
         } 
      glEnd(); 
   } 
   glPopMatrix(); 
} 

extern char loadname[32]; 

void Mod_LoadQ3Model(model_t *mod, void *buffer) 
{ 
   md3header_t *header; 
   md3header_mem_t *mem_head; 
   md3surface_t *surf; 
   md3surface_mem_t *currentsurf; 
   int i, j;//, size, skinnamelen; 
   int posn; 
   char name[128]; 

   byte *cache; 

   //we want to work out the size of the model in memory 
    
   //size of surfaces = surface size + num_shaders * shader_mem size + 
   //      num_triangles * tri size + num_verts * textcoord size + 
   //      num_verts * vert_mem size 
   int surf_size = 0; 
   int mem_size = 0; 

   //pointer to header 
   header = (md3header_t *)buffer; 
   //pointer to the surface list 
   surf = (md3surface_t*)((byte *)buffer + header->surface_offs); 

   surf_size = 0; 
   for (i = 0; i < header->num_surfaces; i++) 
   { 
      surf_size += sizeof(md3surface_mem_t); 
      surf_size += surf->num_surf_shaders * sizeof(md3shader_mem_t); 
      surf_size += surf->num_surf_tris * sizeof(md3tri_mem_t); 
      surf_size += surf->num_surf_verts * sizeof(md3st_mem_t); 
      surf_size += surf->num_surf_verts * surf->num_surf_frames * sizeof(md3vert_mem_t); 

      //goto next surf 
      surf = (md3surface_t*)((byte *)surf + surf->end_offs); 
   } 

   //total size =   header size + num_frames * frame size + num_tags * tag size + 
   //      size of surfaces 
   mem_size = sizeof(md3header_mem_t); 
   mem_size += header->num_frames * sizeof(md3frame_mem_t); 
   mem_size += header->num_tags * sizeof(md3tag_mem_t); 
   mem_size += surf_size; 

   Con_Printf("Loading md3 model...%s (%s)\n", header->filename, mod->name); 

   mem_head = malloc(mem_size); 
   mod->data = mem_head; 

   //stop quake trying to reload md3 models and waste memory 
   //temp = (char *)Cache_Alloc (&mod->cache, 4, loadname); 
   //if (!mod->cache.data) 
   //   return;   //cache alloc failed 

   //sprintf(&chasename[0], "MD3"); 
   //Q_memcpy(temp, &chasename, 4); 


   //setup more mem stuff 
   mod->type = mod_alias; 
   mod->aliastype = MD3IDHEADER; 
   mod->numframes = header->num_frames; 

   //copy stuff across from disk buffer to memory 
   posn = 0; //posn in new buffer 

   //copy header 
   Q_memcpy(mem_head, header, sizeof(md3header_t)); 
   posn += sizeof(md3header_mem_t); 

   //posn points to frames 
   mem_head->frames = (md3frame_mem_t *)((byte *)mem_head + posn); 

   //copy frames 
   Q_memcpy(mem_head->frames, (byte *)header + header->frame_offs, sizeof(md3frame_t)*header->num_frames); 
   posn += sizeof(md3frame_mem_t)*header->num_frames; 

   //posn points to tags 
   mem_head->tags = (md3tag_mem_t *)((byte *)mem_head + posn); 

   //copy tags 
   Q_memcpy(mem_head->tags, (byte *)header + header->tag_offs, sizeof(md3tag_t)*header->num_tags); 
   posn += sizeof(md3tag_mem_t)*header->num_tags; 

   //posn points to surfaces 
   mem_head->surfaces = (md3surface_mem_t *)((byte *)mem_head + posn); 

   //copy surfaces, one at a time 
   surf = (md3surface_t *)((byte *)header + header->surface_offs); 

   currentsurf = mem_head->surfaces; 

   for (i=0; i < header->num_surfaces; i++, currentsurf = currentsurf->nextsurf) 
   { 
      //copy size of surface 
      Q_memcpy((byte *)mem_head + posn, surf, sizeof(md3surface_t)); 
      posn += sizeof(md3surface_mem_t); 

      //posn points to surface->shaders 
      currentsurf->shaders = (md3shader_mem_t *)((byte *)mem_head + posn); 

      //next to have to be redone 
      for (j=0; j < surf->num_surf_shaders; j++){ 
         //copy jth shader accross 
         Q_memcpy((byte *)mem_head + posn, (byte *)surf + surf->shader_offs + j * sizeof(md3shader_t), sizeof(md3shader_t)); 
         posn += sizeof(md3shader_mem_t); //copyed non-mem into mem one 

         //try loading texture here 
         sprintf(&name[0],"override/%s",currentsurf->shaders[j].name); 
         currentsurf->shaders[j].texnum = GL_LoadTexImage(&name[0],true,true); 
         if (currentsurf->shaders[j].texnum == 0){ 
            Con_Printf("Model: %s  Texture missing: %s\n", mod->name, currentsurf->shaders[j].name); 
         } 
      } 
      //posn points to surface->tris 
      currentsurf->tris = (md3tri_mem_t *)((byte *)mem_head + posn); 

      //copy tri 
      Q_memcpy((byte *)mem_head + posn, (byte *)surf + surf->tris_offs, sizeof(md3tri_t) * surf->num_surf_tris); 
      posn += sizeof(md3tri_mem_t) * surf->num_surf_tris; 

      //posn points to surface->tc 
      currentsurf->tc = (md3st_mem_t *)((byte *)mem_head + posn); 

      //copy st 
      Q_memcpy((byte *)mem_head + posn, (byte *)surf + surf->tc_offs, sizeof(md3st_t) * surf->num_surf_verts); 
      posn += sizeof(md3st_t) * surf->num_surf_verts; 

      //posn points to surface->verts 
      currentsurf->verts = (md3vert_mem_t *)((byte *)mem_head + posn); 

      //next to have to be redone 
      for (j=0; j < surf->num_surf_verts; j++){ 
         float lat; 
         float lng; 

         //convert verts from shorts to floats 
         md3vert_mem_t *mem_vert = (md3vert_mem_t *)((byte *)mem_head + posn); 
         md3vert_t *disk_vert = (md3vert_t *)((byte *)surf + surf->vert_offs + j * sizeof(md3vert_t)); 
         mem_vert->vec[0] = (float)disk_vert->vec[0] / 64.0f; 
         mem_vert->vec[1] = (float)disk_vert->vec[1] / 64.0f; 
         mem_vert->vec[2] = (float)disk_vert->vec[2] / 64.0f; 

          
         //work out normals 
         lat = (disk_vert->normal + 255) * (2 * 3.141592654) / 255; 
         lng = ((disk_vert->normal >> 8) & 255) * (2 * 3.141592654) / 255; 
         mem_vert->normal[0] = cos (lat) * sin (lng); 
         mem_vert->normal[1] = sin (lat) * sin (lng); 
         mem_vert->normal[2] = cos (lng); 

         posn += sizeof(md3vert_mem_t); //copyed non-mem into mem one 
      } 

      //point to next surf (or eof) 
      surf = (md3surface_t*)((byte *)surf + surf->end_offs); 

      //if this isnt the last surface then set up next pointer 
      if (i+1 != mem_head->num_surfaces) 
         currentsurf->nextsurf = (md3surface_mem_t *)((byte *)mem_head + posn); 
   } 
   //posn should now equal mem_size 

   VectorCopy(mem_head->frames->mins, mod->mins); 
   VectorCopy(mem_head->frames->maxs, mod->maxs); 

   currentsurf = mem_head->surfaces; 
   for (i = 0; i < mem_head->num_surfaces; i++, currentsurf = currentsurf->nextsurf) 
   { 
      if (*(long *)currentsurf->id != MD3IDHEADER){ 
         Con_Printf("MD3 bad surface for: %s\n", mem_head->filename); 
         free (mem_head); 
         return; 
      } 
   } 

   cache = (byte *)Cache_Alloc (&mod->cache, posn, mod->name); 
   if (!mod->cache.data){ 
      free (mem_head); 
      return;   //cache alloc failed 
   } 

   Q_memcpy(cache, mem_head, posn); 
   free (mem_head); 
} 

#if 0 
int debug = 0; 
multimodel_t *Mod_AddMultiModel (entity_t *entity, model_t *mod); 
void Mod_LoadQ3MultiModel (model_t *mod) 
{ 
   multimodel_t *head, *upper, *lower; 
   int handle; 
   char path[MAX_QPATH]; 
   mod->type = mod_null; 

   sprintf(path, "%sanimation.cfg", mod->name); 
   COM_OpenFile(path, &handle); 

   if(handle) //Q3Player 
   { 
      COM_CloseFile(handle); 
      //Load player 
      lower = Mod_AddMultiModel(NULL, mod);//Allocate a mmodel 
      lower->model = Z_Malloc(sizeof(model_t)); 
      Q_strcpy(lower->model->name, va("%slower.md3", mod->name)); 
      lower->model->needload = TRUE; 
      Mod_LoadModel(lower->model, 1); 
      Q_strcpy(lower->identifier, "lower"); 

      lower->linktype = MULTIMODEL_LINK_STANDARD; 


      upper = Mod_AddMultiModel(NULL, mod);//Allocate a mmodel 
      upper->model = Z_Malloc(sizeof(model_t)); 
      Q_strcpy(upper->model->name, va("%supper.md3", mod->name)); 
      upper->model->needload = TRUE; 
      Mod_LoadModel(upper->model, 1); 
      Q_strcpy(upper->identifier, "upper"); 

      upper->linktype = MULTIMODEL_LINK_TAG; 
      upper->linkedmodel = lower; 
      Q_strcpy(upper->tagname, "tag_torso"); 


      head = Mod_AddMultiModel(NULL, mod);//Allocate a mmodel 
      head->model = Z_Malloc(sizeof(model_t)); 
      Q_strcpy(head->model->name, va("%shead.md3", mod->name)); 
      head->model->needload = TRUE; 
      Mod_LoadModel(head->model, 1); 
      Q_strcpy(head->identifier, "head"); 

      head->linktype = MULTIMODEL_LINK_TAG; 
      head->linkedmodel = upper; 
      Q_strcpy(head->tagname, "tag_head"); 
       
      return; 
   } 

   sprintf(path, "%s", mod->name); 
   Q_strcat(path, Q_strrchr(path, '/')); 
   path[Q_strlen(path)-1] = '\0'; 
   Q_strcat(path, "_hand.md3"); 
   Sys_Error("Trying to find weaponmodel %s", path); 
   COM_OpenFile(path, &handle); 
   if(handle) //W_Weapon 
   { 
      COM_CloseFile(handle); 
   } 
} 
#endif 

#endif 
