/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_main.c

#include "quakedef.h"

#include "gl_md3.h"

entity_t	r_worldentity;

int volumeCmdsBuff[MAX_VOLUME_COMMANDS+128]; //Hack protect against slight overflows
float volumeVertsBuff[MAX_VOLUME_VERTS+128];

qboolean	r_cache_thrash;		// compatability

vec3_t		modelorg, r_entorigin;
entity_t	*currententity;

float color_black[4] = {0.0, 0.0, 0.0, 0.0};
float fog_color[4];

int			shinetex_glass, shinetex_chrome;
int			underwatertexture;

int			r_visframecount;	// bumped when going to a new PVS
int			r_framecount;		// used for dlight push checking

mplane_t	frustum[4];

int			c_brush_polys, c_alias_polys;

qboolean	envmap;				// true during envmap command capture 

qboolean	glare;		// glare stuff - Entar

int			currenttexture = -1;		// to avoid unnecessary texture sets

int			cnttextures[2] = {-1, -1};     // cached

int			particletexture;	// little dot for particles
int			playertextures;		// up to 16 color translated skins

int			mirrortexturenum;	// quake texturenum, not gltexturenum
qboolean	mirror;
mplane_t	*mirror_plane;

vec3_t		lightspot;

// fenix@io.com: model transform interpolation
float old_interpolate_model_transform;

//
// view origin
//
vec3_t	vup;
vec3_t	vpn;
vec3_t	vright;
vec3_t	r_origin;

float	r_world_matrix[16];
float	r_base_world_matrix[16];

//
// screen size info
//
refdef_t	r_refdef;

mleaf_t		*r_viewleaf, *r_oldviewleaf;

texture_t	*r_notexture_mip;

int		d_lightstylevalue[256];	// 8.8 fraction of base light value


void R_MarkLeaves (void);

cvar_t	r_norefresh = {"r_norefresh","0"};
cvar_t	r_drawentities = {"r_drawentities","1"};
cvar_t	r_drawviewmodel = {"r_drawviewmodel","1"};
cvar_t	r_speeds = {"r_speeds","0"};
cvar_t	r_fullbright = {"r_fullbright","0"};
cvar_t	r_lightmap = {"r_lightmap","0"};
cvar_t	r_shadows = {"r_shadows","1"};

cvar_t gl_asynch = {"gl_asynch","1"}; //interpolation stuff
cvar_t r_interpolations = {"r_interpolations","6"}; //more interpolation stuff // was 10, made 6 by Entar
//cvar_t  r_interpolate_model_animation = { "r_interpolate_model_animation", "1", true }; // guess :D
cvar_t  r_interpolate_model_a = { "r_interpolate_model_a", "1", true }; // guess :D
//cvar_t  r_interpolate_model_transform = { "r_interpolate_model_transform", "1", true }; // Yup...
cvar_t  r_interpolate_model_t = { "r_interpolate_model_t", "1", true }; // Yup...

cvar_t	r_mirroralpha = {"r_mirroralpha","1"};
cvar_t	r_wateralpha = {"r_wateralpha","1"};
cvar_t	r_dynamic = {"r_dynamic","1"};
cvar_t	r_novis = {"r_novis","0"};

cvar_t	gl_finish = {"gl_finish","0"};
cvar_t	gl_clear = {"gl_clear","0"};
cvar_t	gl_cull = {"gl_cull","1"};
cvar_t	gl_texsort = {"gl_texsort","1"};
cvar_t	gl_smoothmodels = {"gl_smoothmodels","1"};
cvar_t	gl_affinemodels = {"gl_affinemodels","0"};
cvar_t	gl_polyblend = {"gl_polyblend","1"};
cvar_t	gl_flashblend = {"gl_flashblend","0"};
cvar_t	gl_playermip = {"gl_playermip","0"};
cvar_t	gl_nocolors = {"gl_nocolors","0"};
cvar_t	gl_keeptjunctions = {"gl_keeptjunctions","1", true};
cvar_t	gl_reporttjunctions = {"gl_reporttjunctions","0"};
cvar_t	gl_doubleeyes = {"gl_doubleeys", "1"};

cvar_t	sh_glares = {"sh_glares", "1", true};
cvar_t	gl_wireframe = {"gl_wireframe", "1", true};
cvar_t	cl_sbar = {"cl_sbar", "1", true};
cvar_t	gl_dark = {"gl_dark","1", true};

cvar_t  sh_worldshadows = {"sh_worldshadows","1"};//PENTA: brushes cast shadows on/off

//cvar_t   *r_glow_blur_steps; 
//cvar_t   *r_glow_darken_steps; 
//cvar_t   *r_glow_strength;

//r_postprocess.cpp 
cvar_t  r_glow_blur_steps = {"r_glow_blur_steps", "4", true}; 
cvar_t  r_glow_darken_steps = {"r_glow_darken_steps", "3", true}; 
cvar_t  r_glow_strength = {"r_glow_strength", "2", true};

/*
cvar_t	shinetex_glass = {"shinetex_glass", "1", true};;
cvar_t	shinetex_chrome = {"shinetex_chrome", "1", true};;
cvar_t	GL_TEXTURE0_ARB = {"GL_TEXTURE0_ARB", "1", true};;
cvar_t	SURF_SHINY_GLASS = {"SURF_SHINY_GLASS", "1", true};;
cvar_t	SURF_SHINY_METAL = {"SURF_SHINY_METAL", "1", true};;
cvar_t	gl_detail = {"gl_detail", "1", true};;
cvar_t	gl_textureunits = {"gl_texutreunits", "1", true};
cvar_t	r_outline = {"r_outline", "1", true};
cvar_t	gl_shiny = {"gl_shiny","1", true};
*/
cvar_t	gl_caustics = {"gl_caustics", "1", true};
cvar_t  gl_shiny = {"gl_shiny", "1", true};
cvar_t	gl_detail = {"gl_detail","1", true};
cvar_t	gl_dualwater = {"gl_dualwater","1", true};
cvar_t	gl_watershader = {"gl_watershader","1"};//PENTA: water shaders ON/OFF
//cvar_t  gl_rainparticles = {"gl_rainparticles", "1"};
//cvar_t	gl_rainparticles_quality = {"gl_rainparticles_quality", "1"};
cvar_t	gl_particles = {"gl_particles", "0", true};

cvar_t	sh_colormaps = {"sh_colormaps","1"};//PENTA: enable disable textures on the world (light will remain)
cvar_t	sh_noshadowpopping = {"sh_noshadowpopping","1"};

//cvar_t	sbar_alpha	= {"sbar_alpha", "0", true}; // Tomaz - Sbar Alpha

extern	cvar_t	gl_ztrick;

extern	cvar_t		v_gamma; // muff

// idea originally nicked from LordHavoc 
// re-worked + extended - muff 5 Feb 2001
// called inside polyblend

void DoGamma()
{
	if (v_gamma.value < 0.2)
		v_gamma.value = 0.2;
	if (v_gamma.value >= 1)
	{
		v_gamma.value = 1;
		return;
	}

	//believe it or not this actually does brighten the picture!!
	glBlendFunc (  GL_DST_COLOR, GL_ONE_MINUS_SRC_ALPHA);
	glColor4f (1, 1, 1, v_gamma.value);

	glBegin (GL_QUADS);

	glVertex3f (10, 100, 100);
	glVertex3f (10, -100, 100);
	glVertex3f (10, -100, -100);
	glVertex3f (10, 100, -100);

	//if we do this twice, we double the brightening effect for a wider range of gamma's
	glVertex3f (11, 100, 100);
	glVertex3f (11, -100, 100);
	glVertex3f (11, -100, -100);
	glVertex3f (11, 100, -100);

	glEnd ();
}

/*
=================
R_CullBox

Returns true if the box is completely outside the frustom
=================
*/
qboolean R_CullBox (vec3_t mins, vec3_t maxs)
{
	int		i;

	for (i=0 ; i<4 ; i++)
		if (BoxOnPlaneSide (mins, maxs, &frustum[i]) == 2)
			return true;
	return false;
}


void R_RotateForEntity (entity_t *e)
{
    glTranslatef (e->origin[0],  e->origin[1],  e->origin[2]);

    glRotatef (e->angles[1],  0, 0, 1);
    glRotatef (-e->angles[0],  0, 1, 0);
    glRotatef (e->angles[2],  1, 0, 0);
}

/*
=============
R_BlendedRotateForEntity

fenix@io.com: model transform interpolation
=============
*/
//added , int shadow
void R_BlendedRotateForEntity (entity_t *e, int shadow)
{
    float blend;
    vec3_t d;
    int i;

    // positional interpolation

    if (e->translate_start_time == 0)
    {
        e->translate_start_time = realtime;
        VectorCopy (e->origin, e->origin1);
        VectorCopy (e->origin, e->origin2);
    }

    if (!VectorCompare (e->origin, e->origin2))
    {
        e->translate_start_time = realtime;
        VectorCopy (e->origin2, e->origin1);
        VectorCopy (e->origin,  e->origin2);
        blend = 0;
    }
    else
    {
        blend = (realtime - e->translate_start_time) / 0.1;

        if (cl.paused || blend > 1) blend = 1;
    }
    VectorSubtract (e->origin2, e->origin1, d);

    glTranslatef (
        e->origin1[0] + (blend * d[0]),
        e->origin1[1] + (blend * d[1]),
        e->origin1[2] + (blend * d[2]));

    // orientation interpolation (Euler angles, yuck!)

    if (e->rotate_start_time == 0)
    {
        e->rotate_start_time = realtime;
        VectorCopy (e->angles, e->angles1);
        VectorCopy (e->angles, e->angles2);
    }

    if (!VectorCompare (e->angles, e->angles2))
    {
        e->rotate_start_time = realtime;
        VectorCopy (e->angles2, e->angles1);
        VectorCopy (e->angles,  e->angles2);
        blend = 0;
    }
    else
    {
        blend = (realtime - e->rotate_start_time) / 0.1;

        if (cl.paused || blend > 1) blend = 1;
    }

    VectorSubtract (e->angles2, e->angles1, d);

    // always interpolate along the shortest path
    for (i = 0; i < 3; i++) 
    {
        if (d[i] > 180)
        {
            d[i] -= 360;
        }
        else if (d[i] < -180)
        {
            d[i] += 360;
        }
    }

    //glRotatef ( e->angles1[1] + ( blend * d[1]),  0, 0, 1);
    //glRotatef (-e->angles1[0] + (-blend * d[0]),  0, 1, 0);
    //glRotatef ( e->angles1[2] + ( blend * d[2]),  1, 0, 0);
	glRotatef ( e->angles1[1] + ( blend * d[1]),  0, 0, 1);

	// muff@yakko.globalnet.co.uk - 05 Mar 00
	// added to allow shadows to behave properly
	// if shadow it only rotates in horizontal plane
	// will need updating when slope stuff is in place
	if (shadow==0)
	{
			glRotatef (-e->angles1[0] + (-blend * d[0]),  0, 1, 0); // this rotates tilt
			glRotatef ( e->angles1[2] + ( blend * d[2]),  1, 0, 0); // this is the one that causes the weird missile stuff
	}

}

/*
=============================================================

  SPRITE MODELS

=============================================================
*/

/*
================
R_GetSpriteFrame
================
*/
mspriteframe_t *R_GetSpriteFrame (entity_t *currententity)
{
	msprite_t		*psprite;
	mspritegroup_t	*pspritegroup;
	mspriteframe_t	*pspriteframe;
	int				i, numframes, frame;
	float			*pintervals, fullinterval, targettime, time;

	psprite = currententity->model->cache.data;
	frame = currententity->frame;

	if ((frame >= psprite->numframes) || (frame < 0))
	{
		Con_Printf ("R_DrawSprite: no such frame %d\n", frame);
		frame = 0;
	}

	if (psprite->frames[frame].type == SPR_SINGLE)
	{
		pspriteframe = psprite->frames[frame].frameptr;
	}
	else
	{
		pspritegroup = (mspritegroup_t *)psprite->frames[frame].frameptr;
		pintervals = pspritegroup->intervals;
		numframes = pspritegroup->numframes;
		fullinterval = pintervals[numframes-1];

		time = cl.time + currententity->syncbase;

	// when loading in Mod_LoadSpriteGroup, we guaranteed all interval values
	// are positive, so we don't have to worry about division by 0
		targettime = time - ((int)(time / fullinterval)) * fullinterval;

		for (i=0 ; i<(numframes-1) ; i++)
		{
			if (pintervals[i] > targettime)
				break;
		}

		pspriteframe = pspritegroup->frames[i];
	}

	return pspriteframe;
}


/*
=================
R_DrawSpriteModel

=================
*/
void R_DrawSpriteModel (entity_t *e)
{
	vec3_t	point;
	mspriteframe_t	*frame;
	float		*up, *right;
	vec3_t		v_forward, v_right, v_up;
	msprite_t		*psprite;

	glDisable(GL_ALPHA_TEST);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	// don't even bother culling, because it's just a single
	// polygon without a surface cache
	frame = R_GetSpriteFrame (e);
	psprite = currententity->model->cache.data;

	if (psprite->type == SPR_ORIENTED)
	{	// bullet marks on walls
		AngleVectors (currententity->angles, v_forward, v_right, v_up);
		up = v_up;
		right = v_right;
	}
	else
	{	// normal sprite
		up = vup;
		right = vright;
	}

	glColor3f (1,1,1);

	GL_DisableMultitexture();

    GL_Bind(frame->gl_texturenum);

	glEnable (GL_ALPHA_TEST);
	glBegin (GL_QUADS);

	glTexCoord2f (0, 1);
	VectorMA (e->origin, frame->down, up, point);
	VectorMA (point, frame->left, right, point);
	glVertex3fv (point);

	glTexCoord2f (0, 0);
	VectorMA (e->origin, frame->up, up, point);
	VectorMA (point, frame->left, right, point);
	glVertex3fv (point);

	glTexCoord2f (1, 0);
	VectorMA (e->origin, frame->up, up, point);
	VectorMA (point, frame->right, right, point);
	glVertex3fv (point);

	glTexCoord2f (1, 1);
	VectorMA (e->origin, frame->down, up, point);
	VectorMA (point, frame->right, right, point);
	glVertex3fv (point);
	
	glEnd ();

	glDisable (GL_ALPHA_TEST);

	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glDisable(GL_BLEND);
	glDisable(GL_ALPHA_TEST);

}

/*
=============================================================

  ALIAS MODELS

=============================================================
*/


#define NUMVERTEXNORMALS	162

float	r_avertexnormals[NUMVERTEXNORMALS][3] = {
#include "anorms.h"
};

vec3_t	shadevector;
//float	shadelight, ambientlight; // LordHavoc: .lit support, removed shadelight, ambientlight

// precalculated dot products for quantized angles
#define SHADEDOT_QUANT 16
float	r_avertexnormal_dots[SHADEDOT_QUANT][256] =
#include "anorm_dots.h"
;

float	*shadedots = r_avertexnormal_dots[0];

// fenix@io.com: model animation interpolation
int lastposenum0;

int	lastposenum;
/*================*/
   extern vec3_t lightcolor; // LH
//void GL_DrawAliasInterpolatedFrame (aliashdr_t *paliashdr, int posenum, int oldposenum, int interp) 
void GL_DrawAliasInterpolatedFrame (aliashdr_t *paliashdr, int posenum, int oldposenum, int interp, float alpha)
{ 
   float s, t; 
   float  l; 
   float interpolations; 
   int  i, j; 
   int  index; 
   trivertx_t *v, *verts, *oldverts; 
   int  list; 
   int  *order; 
   vec3_t point; 
   float *normal; 
   int  count; 
   
   //Con_SafePrintf("alpha is %f\n", alpha);

//   glEnable (GL_BLEND);
//   glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    if (alpha != 1) 
    { 
       glEnable (GL_BLEND); 
       //glDepthMask (GL_TRUE); 
	   glDepthMask (GL_FALSE); 
    } 
    else 
    { 
       glDisable (GL_BLEND); 
       glDepthMask (GL_TRUE); // was false, now fixed - DP
    } 

   lastposenum = posenum;
   interpolations = interp/r_interpolations.value; 
   verts = (trivertx_t *)((byte *)paliashdr + paliashdr->posedata); 
   oldverts = verts; 
   verts += posenum * paliashdr->poseverts; 
   if (oldposenum >= 0) 
      oldverts += oldposenum * paliashdr->poseverts; 
   else 
      oldverts += posenum * paliashdr->poseverts; 

   order = (int *)((byte *)paliashdr + paliashdr->commands); 

   while (1) 
      { 
      // get the vertex count and primitive type 
      count = *order++; 
      if (!count) 
         break;  // done 
      if (count < 0) 
         { 
         count = -count; 
         glBegin (GL_TRIANGLE_FAN); 
         } 
      else 
         glBegin (GL_TRIANGLE_STRIP); 
      do 
         { 
		 // texture coordinates come from the draw list 
         glTexCoord2f (((float *)order)[0], ((float *)order)[1]); 
         order += 2; 
         // normals and vertexes come from the frame list 
         //l = shadedots[verts->lightnormalindex] * shadelight; 
		 l = shadedots[verts->lightnormalindex];
		 //glColor4f (1,1,1,0.5);
		 //glColor4f (1,1,1,alpha);
		 glColor4f (l * lightcolor[0]+0.01, l * lightcolor[1]+0.01, l * lightcolor[2]+0.01, alpha);		 
		 glVertex3f (oldverts->v[0] + ((verts->v[0] - oldverts->v[0])*interpolations), 
                     oldverts->v[1] + ((verts->v[1] - oldverts->v[1])*interpolations),  
                     oldverts->v[2] + ((verts->v[2] - oldverts->v[2])*interpolations)); 
         verts++; 
         oldverts++; 
         } while (--count); 
      glEnd (); 
      }
   glColor4f (1,1,1,1);
   glDepthMask(GL_TRUE);
   glDisable(GL_BLEND);
} 

/*
=============
GL_DrawAliasFrame
=============
*/
void GL_DrawAliasFrame (aliashdr_t *paliashdr, int posenum, float alpha)
{
	float 	l;
	trivertx_t	*verts;
	int		*order;
	int		count;
	extern vec3_t lightcolor; // LordHavoc: .lit support to the definitions at the top

	lastposenum = posenum;
	// Thanks LH :)
	/*
	glEnable (GL_BLEND); 
	//GL_TexEnv(GL_MODULATE); 
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	*/
	
    if (alpha != 1) 
    { 
       glEnable (GL_BLEND); 
       //glDepthMask (GL_TRUE); 
	   glDepthMask (GL_FALSE); 
    } 
    else 
    { 
       glDisable (GL_BLEND); 
       glDepthMask (GL_TRUE); // was false, now fixed - LH
    } 
	// End LH :)
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE); 
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); 
	verts = (trivertx_t *)((byte *)paliashdr + paliashdr->posedata);
	verts += posenum * paliashdr->poseverts;
	order = (int *)((byte *)paliashdr + paliashdr->commands);

	while (1)
	{
		// get the vertex count and primitive type
		count = *order++;
		if (!count)
			break;		// done
		if (count < 0)
		{
			count = -count;
			glBegin (GL_TRIANGLE_FAN);
		}
		else
			glBegin (GL_TRIANGLE_STRIP);

		do
		{
			// texture coordinates come from the draw list
			glTexCoord2f (((float *)order)[0], ((float *)order)[1]);
			order += 2;

			// normals and vertexes come from the frame list
			//l = shadedots[verts->lightnormalindex] * shadelight; // LH - old code
			//glColor3f (l, l, l); // LH - old code
			l = shadedots[verts->lightnormalindex];
			//glColor3f (l * lightcolor[0], l * lightcolor[1], l * lightcolor[2]);
			glColor4f (l * lightcolor[0], l * lightcolor[1], l * lightcolor[2], alpha);
			glVertex3f (verts->v[0], verts->v[1], verts->v[2]);
			verts++;
		} while (--count);
	
		glEnd ();
	}
	glDepthMask(GL_TRUE);
	glDisable(GL_BLEND);
}

/*
=============
GL_DrawAliasBlendedFrame

fenix@io.com: model animation interpolation
=============
*/
void GL_DrawAliasBlendedFrame (aliashdr_t *paliashdr, int pose1, int pose2, float blend)
{
   float       l;
   trivertx_t* verts1;
   trivertx_t* verts2;
   int*        order;
   int         count;
   vec3_t      d;

   lastposenum0 = pose1;
   lastposenum  = pose2;

   verts1  = (trivertx_t *)((byte *)paliashdr + paliashdr->posedata);
   verts2  = verts1;

   verts1 += pose1 * paliashdr->poseverts;
   verts2 += pose2 * paliashdr->poseverts;

   order = (int *)((byte *)paliashdr + paliashdr->commands);
   
   for (;;)
      {
      // get the vertex count and primitive type
      count = *order++;

      if (!count) break;

      if (count < 0)
         {
         count = -count;
         glBegin (GL_TRIANGLE_FAN);
         }
      else
         {
         glBegin (GL_TRIANGLE_STRIP);
         }

      do
         {
         // texture coordinates come from the draw list
         glTexCoord2f (((float *)order)[0], ((float *)order)[1]);
         order += 2;

         // normals and vertexes come from the frame list
         // blend the light intensity from the two frames together
         d[0] = shadedots[verts2->lightnormalindex] -
                shadedots[verts1->lightnormalindex];

         //l = shadelight * (shadedots[verts1->lightnormalindex] + (blend * d[0]));
		 l = (shadedots[verts1->lightnormalindex] + (blend * d[0]));
         glColor3f (l, l, l);

         VectorSubtract(verts2->v, verts1->v, d);

         // blend the vertex positions from each frame together
         glVertex3f (
                    verts1->v[0] + (blend * d[0]),
                    verts1->v[1] + (blend * d[1]),
                    verts1->v[2] + (blend * d[2]));

         verts1++;
         verts2++;
         } while (--count);
      glEnd ();
      }
}


/*
=============
GL_DrawAliasShadow
=============

extern	vec3_t			lightspot;

void GL_DrawAliasShadow (aliashdr_t *paliashdr, int posenum)
{
	trivertx_t	*verts;
	int		*order;
	vec3_t	point;
	float	height, lheight;
	int		count;

	lheight = currententity->origin[2] - lightspot[2];

	height = 0;
	verts = (trivertx_t *)((byte *)paliashdr + paliashdr->posedata);
	verts += posenum * paliashdr->poseverts;
	order = (int *)((byte *)paliashdr + paliashdr->commands);

	height = -lheight + 1.0;

	while (1)
	{
		// get the vertex count and primitive type
		count = *order++;
		if (!count)
			break;		// done
		if (count < 0)
		{
			count = -count;
			glBegin (GL_TRIANGLE_FAN);
		}
		else
			glBegin (GL_TRIANGLE_STRIP);

		do
		{
			// texture coordinates come from the draw list
			// (skipped for shadows) glTexCoord2fv ((float *)order);
			order += 2;

			// normals and vertexes come from the frame list
			point[0] = verts->v[0] * paliashdr->scale[0] + paliashdr->scale_origin[0];
			point[1] = verts->v[1] * paliashdr->scale[1] + paliashdr->scale_origin[1];
			point[2] = verts->v[2] * paliashdr->scale[2] + paliashdr->scale_origin[2];

			point[0] -= shadevector[0]*(point[2]+lheight);
			point[1] -= shadevector[1]*(point[2]+lheight);
			point[2] = height;
//			height -= 0.001;
			glVertex3fv (point);

			verts++;
		} while (--count);

		glEnd ();
	}	
}
*/
void GL_DrawAliasShadow (aliashdr_t *paliashdr, int posenum)
{
	float	s, t, l;
	int		i, j;
	int		index;
	trivertx_t	*v, *verts;
	int		list;
	int		*order;
	vec3_t	point;
	float	*normal;
	float	height, lheight;
	int		count;

	// muff@yakko.globalnet.co.uk - 17 Mar 2000
	trace_t		steptrace, downtrace;
	vec3_t		dest,stop,downmove;
	float		s1,c1; //this will store the sin and cos values as they are used so heavily
	//end of muff

	lheight = currententity->origin[2] - lightspot[2];

	height = 0;
	verts = (trivertx_t *)((byte *)paliashdr + paliashdr->posedata);
	verts += posenum * paliashdr->poseverts;
	order = (int *)((byte *)paliashdr + paliashdr->commands);

	height = -lheight ;//+ 1.0; //was 1.0 - muff

	// muff@yakko.globalnet.co.uk - 17 Mar 2000
	// better shadowing, now takes angle of ground into account	
	// cast a traceline into the floor directly below the player
	// and gets normals from this
	VectorCopy (currententity->origin, downmove);
	downmove[2] = downmove[2] - 4096;
	memset (&downtrace, 0, sizeof(downtrace));
	SV_RecursiveHullCheck (cl.worldmodel->hulls, 0, 0, 1, currententity->origin, downmove, &downtrace);

	// calculate the all important angles to keep speed up
	s1 = sin( currententity->angles[1]/180*M_PI);
	c1 = cos( currententity->angles[1]/180*M_PI);

	// end of muff

	while (1)
	{
		// get the vertex count and primitive type
		count = *order++;
		if (!count)
			break;		// done
		if (count < 0)
		{
			count = -count;
			glBegin (GL_TRIANGLE_FAN);
		}
		else
			glBegin (GL_TRIANGLE_STRIP);

		do
		{
			// texture coordinates come from the draw list
			// (skipped for shadows) glTexCoord2fv ((float *)order);
			order += 2;

			// normals and vertexes come from the frame list
			point[0] = verts->v[0] * paliashdr->scale[0] + paliashdr->scale_origin[0];
			point[1] = verts->v[1] * paliashdr->scale[1] + paliashdr->scale_origin[1];
			point[2] = verts->v[2] * paliashdr->scale[2] + paliashdr->scale_origin[2];

			// muff@yakko.globalnet.co.uk - 17 Mar 2000
			// first remove the existing height adjustment, as we are going to do this ourselves
			//original code
			// point[0] -= shadevector[0]*(point[0] +lheight);
			// point[1] -= shadevector[1]*(point[1] +lheight);
			//	point[2] = height;
			//	height -= 0.001;	
			// end of original code

			point[0] -= shadevector[0]*(point[0]); 
			point[1] -= shadevector[1]*(point[1]);
			point[2] -= shadevector[2]*(point[2]);

			//drop it down to floor
			point[2] = point[2] - (currententity->origin[2] - downtrace.endpos[2]) ;

			// now adjust the point with respect to all the normals of the tracepoint
			// I'm probably missing an easy piece of maths to do this
			// it reads messy, but works ;)
			point[2] +=	 (    	(point[1] * (s1 * downtrace.plane.normal[0])) -
					     	(point[0] * (c1 * downtrace.plane.normal[0])) -
						(point[0] * (s1 * downtrace.plane.normal[1])) - 
						(point[1] * (c1 * downtrace.plane.normal[1])) 
					 	 ) +  ((1.0 - downtrace.plane.normal[2])*20)
						 + 0.2 ;

		        //end of muff

			glVertex3fv (point);

			verts++;
		} while (--count);

		glEnd ();
	}	
}

void R_SetupAliasInterpolatedFrame (int frame, int lastframe, float interp, aliashdr_t *paliashdr, float alpha) 
{ 
   int   pose, numposes, oldpose; 
   float   interval; 
   if ((frame >= paliashdr->numframes) || (frame < 0)) 
      { 
      Con_DPrintf ("R_AliasSetupFrame: no such frame %d\n", frame); 
      frame = 0; 
      } 
   if ((lastframe >= paliashdr->numframes) || (lastframe < 0)) 
      { 
      Con_DPrintf ("R_AliasSetupFrame: no such last frame %d\n", lastframe); 
      lastframe = 0; 
      } 
   pose = paliashdr->frames[frame].firstpose; 
   numposes = paliashdr->frames[frame].numposes; 
   if (numposes > 1) 
      { 
      interval = paliashdr->frames[frame].interval; 
      pose += (int)(cl.time / interval) % numposes; 
      } 
   oldpose = paliashdr->frames[lastframe].firstpose; 
   numposes = paliashdr->frames[lastframe].numposes; 
   if (numposes > 1) 
      { 
      interval = paliashdr->frames[lastframe].interval; 
      oldpose += (int)(cl.time / interval) % numposes; 
      } 
   //GL_DrawAliasInterpolatedFrame (paliashdr, pose, oldpose, interp); 
   GL_DrawAliasInterpolatedFrame (paliashdr, pose, oldpose, interp, alpha); 
} 

#define PROJECTION_DISTANCE			10
#define MAX_STENCIL_ENTS			1024
static entity_t *g_stencilEnts[MAX_STENCIL_ENTS];
static int g_numStencilEnts = 0;

/*
=============
RotateLightVector
- thanks Spike :)
=============
*/

void RotateLightVector(vec3_t angles, vec3_t origin, vec3_t lightpoint, vec3_t result) 
{ 
vec3_t f, r, u, offs; 

angles[0]*=-1; 
AngleVectors(angles, f, r, u); 
angles[0]*=-1; 

offs[0] = lightpoint[0] - origin[0]; 
offs[1] = lightpoint[1] - origin[1]; 
offs[2] = lightpoint[2] - origin[2]; 

result[0] = DotProduct (offs, f); 
result[1] = -DotProduct (offs, r); 
result[2] = DotProduct (offs, u); 
} 

/*
=============
GL_GetTriNeighbors

Get the neighbors for a given triangle (meaning, said neighbor
shares an edge with this triangle).
-rww
=============
*/
void GL_GetTriNeighbors(aliashdr_t *paliashdr, mtriangle_t *tris, mtriangle_t *tri, int *neighbors)
{
	static mtriangle_t *n;
	int i = 0;
	int numN = 0;
	static int j, k;
	static int vertA1, vertA2, vertB1, vertB2;

	neighbors[0] = neighbors[1] = neighbors[2] = -1;

	while (i < paliashdr->numtris)
	{
		n = &tris[i];

		if (n != tri)
		{
			j = 0;
			while (j < 3)
			{
				vertA1 = tri->vertindex[j];
				vertA2 = tri->vertindex[(j+1)%3];

				k = 0;
				while (k < 3)
				{
					vertB1 = n->vertindex[k];
					vertB2 = n->vertindex[(k+1)%3];

					if ((vertA1 == vertB1 && vertA2 == vertB2) || (vertA1 == vertB2 && vertA2 == vertB1))
					{
						if (neighbors[j] == -1)
						{
							neighbors[j] = i;
						}
						else
						{ //this edge is shared by more than 2 tris apparently.
						  //this seems to actually happen at times, which is why
						  //we are bothering to check for it.
							neighbors[j] = -2;
						}
					}

					k++;
				}

				j++;
			}
		}

		i++;
	}

	j = 0;
	while (j < 3)
	{ //mark any -2 as -1 now since we're done
		if (neighbors[j] == -2)
		{
			neighbors[j] = -1;
		}
		j++;
	}
}

/*
=============
GL_TriPrecalc

Do whatever precalc'ing we can (in this case just get neighbors).
We could also pre-determine planeEq's for each tri*each frame in
here if we desired.
-rww
=============
*/
void GL_TriPrecalc(aliashdr_t *paliashdr, mtriangle_t *tris, trivertx_t *verts)
{
	int i = 0;
	mtriangle_t *tri;

	while (i < paliashdr->numtris)
	{
		tri = &tris[i];
		GL_GetTriNeighbors(paliashdr, tris, tri, tri->neighbors);
		i++;
	}
}

/*
=============
GL_TriFacingLight

Determine of a given triangle is facing the light position. A lot
of this could probably be precalculated, but I am lazy. To store
the planeEq we would have to store one for every frame since the
tri remains the same but the verts change based on the frame.
The calculation is relatively inexpensive so it is at least
somewhat reasonable to perform it in realtime.
-rww
=============
*/
int GL_TriFacingLight(mtriangle_t *tri, trivertx_t *verts, vec3_t lightPosition)
{
	byte *v1;
	byte *v2;
	byte *v3;
	float side;
	float planeEq[4];

	v1 = &verts[tri->vertindex[0]].v[0];
	v2 = &verts[tri->vertindex[1]].v[0];
	v3 = &verts[tri->vertindex[2]].v[0];

	planeEq[0] = v1[1]*(v2[2]-v3[2]) + v2[1]*(v3[2]-v1[2]) + v3[1]*(v1[2]-v2[2]);
	planeEq[1] = v1[2]*(v2[0]-v3[0]) + v2[2]*(v3[0]-v1[0]) + v3[2]*(v1[0]-v2[0]);
	planeEq[2] = v1[0]*(v2[1]-v3[1]) + v2[0]*(v3[1]-v1[1]) + v3[0]*(v1[1]-v2[1]);
	planeEq[3] = -( v1[0]*( v2[1]*v3[2] - v3[1]*v2[2] ) +
				v2[0]*(v3[1]*v1[2] - v1[1]*v3[2]) +
				v3[0]*(v1[1]*v2[2] - v2[1]*v1[2]) );

	side = planeEq[0]*lightPosition[0]+
		planeEq[1]*lightPosition[1]+
		planeEq[2]*lightPosition[2]+
		planeEq[3];

	if ( side > 0 )
	{
		return 1;
	}

	return 0;
}

/*
=============
GL_ShadowPass

Determine and draw our shadow volume
-rww
=============
*/
void GL_ShadowPass(aliashdr_t *paliashdr, int posenum, vec3_t lightPosition)
{
	static byte *v1;
	static byte *v2;
	static vec3_t v3;
	static vec3_t v4;
	static int i;
	static int j;
	static int neighborIndex;
	static trivertx_t *verts;
	static mtriangle_t *tris;

	verts = (trivertx_t *)((byte *)paliashdr + paliashdr->baseposedata);
	tris = (mtriangle_t *)((byte *)paliashdr + paliashdr->triangles);
	verts += posenum * paliashdr->numverts;

	//go through all the triangles
	for ( i = 0; i < paliashdr->numtris; i++ )
	{
		mtriangle_t *tri = &tris[i];

		//Only bother if this tri is facing the light pos
		if (GL_TriFacingLight(tri, verts, lightPosition))
		{
			for ( j = 0; j < 3; j++ )
			{
				neighborIndex = tri->neighbors[j];

				//If the tri has no neighbor or the neighbor is not facing the light,
				//then it is an edge
				if (neighborIndex == -1 || !GL_TriFacingLight(&tris[neighborIndex], verts, lightPosition))
				{
					v1 = &verts[tri->vertindex[j]].v[0];
					v2 = &verts[tri->vertindex[( j+1 )%3]].v[0];

					//get positions of v3 and v4 based on the light position
					v3[0] = ( v1[0]-lightPosition[0] )*PROJECTION_DISTANCE;
					v3[1] = ( v1[1]-lightPosition[1] )*PROJECTION_DISTANCE;
					v3[2] = ( v1[2]-lightPosition[2] )*PROJECTION_DISTANCE;

					v4[0] = ( v2[0]-lightPosition[0] )*PROJECTION_DISTANCE;
					v4[1] = ( v2[1]-lightPosition[1] )*PROJECTION_DISTANCE;
					v4[2] = ( v2[2]-lightPosition[2] )*PROJECTION_DISTANCE;

					//Now draw the quad from the two verts to the projected light
					//verts
					glBegin( GL_QUAD_STRIP );
						glVertex3f( v1[0], v1[1], v1[2] );
						glVertex3f( v1[0]+v3[0], v1[1]+v3[1], v1[2]+v3[2] );
						glVertex3f( v2[0], v2[1], v2[2] );
						glVertex3f( v2[0]+v4[0], v2[1]+v4[1], v2[2]+v4[2] );
					glEnd();
				}
			}
			//draw the front cap for the shadow volume
			glBegin(GL_TRIANGLES);
			for ( j = 0; j < 3; j++ )
			{
				v1 = &verts[tri->vertindex[j]].v[0];
				glVertex3f(v1[0], v1[1], v1[2]);
			}
			glEnd();

			//Now the back cap. We draw it backwards to assure it is drawn
			//facing outward from the shadow volume.
			glBegin(GL_TRIANGLES);
			for (j = 0; j < 3; j++ )
			{
				v1 = &verts[tri->vertindex[2-j]].v[0];
				v3[0] = ( v1[0]-lightPosition[0] )*PROJECTION_DISTANCE;
				v3[1] = ( v1[1]-lightPosition[1] )*PROJECTION_DISTANCE;
				v3[2] = ( v1[2]-lightPosition[2] )*PROJECTION_DISTANCE;

				glVertex3f( v1[0]+v3[0], v1[1]+v3[1], v1[2]+v3[2] );
			}
			glEnd();

		}
	}
}

/*
=============
GL_DrawAliasStencilShadow

Do both of the stencil buffer passes
-rww
=============
*/
void GL_DrawAliasStencilShadow (aliashdr_t *paliashdr, int posenum, vec3_t lightPosition)
{
	/*
	glCullFace(GL_BACK);
	glStencilOp( GL_KEEP, GL_KEEP, GL_INCR );
	GL_ShadowPass( paliashdr, posenum, lightPosition );

	glCullFace(GL_FRONT);
	glStencilOp( GL_KEEP, GL_KEEP, GL_DECR );
	GL_ShadowPass( paliashdr, posenum, lightPosition );
	*/
	glCullFace(GL_FRONT);
	glStencilOp( GL_KEEP, GL_INCR, GL_KEEP );
	GL_ShadowPass( paliashdr, posenum, lightPosition );

	glCullFace(GL_BACK);
	glStencilOp( GL_KEEP, GL_DECR, GL_KEEP );
	GL_ShadowPass( paliashdr, posenum, lightPosition );

}

/*
=================
GL_StencilShadowModel

Do the shadowing for an ent
-rww
=================
*/
void GL_StencilShadowModel(entity_t *e)
{

	int				pose, numposes;
	float			interval;
	float			len;
	aliashdr_t		*paliashdr;
	//vec3_t			lightPos;
	vec3_t			lightPos, relativelightpos;
	vec3_t			fwd, right, up, ang, ang2;
	//model_t  *clmodel;
	//int				*clmodel;

    //clmodel = currententity->model; 

	if (e == &cl.viewent)
	{
		return;
	}

	if (!e->model)
	{
		return;
	}

	//if (!strcmp (clmodel->name, "progs/flame2.mdl"))
	if (!strcmp(currententity->model->name, "progs/flame.mdl"))
	{
		return;
	}

	paliashdr = (aliashdr_t *)Mod_Extradata (e->model);

	if ((e->frame >= paliashdr->numframes) || (e->frame < 0))
	{
		Con_DPrintf ("R_AliasSetupFrame: no such frame %d\n", e->frame);
		e->frame = 0;
	}

	pose = paliashdr->frames[e->frame].firstpose;
	numposes = paliashdr->frames[e->frame].numposes;

	if (numposes > 1)
	{
		interval = paliashdr->frames[e->frame].interval;
		pose += (int)(cl.time / interval) % numposes;
	}

	//Just some random light position.
	//This is relative to the entity position and angles.
	lightPos[0] = 0.0f;
	lightPos[1] = 0.0f;
	lightPos[2] = 512.0f;

	glPushMatrix ();
	R_RotateForEntity (e);
	glTranslatef (paliashdr->scale_origin[0], paliashdr->scale_origin[1], paliashdr->scale_origin[2]);
	glScalef (paliashdr->scale[0], paliashdr->scale[1], paliashdr->scale[2]);
	//GL_DrawAliasStencilShadow(paliashdr, pose, lightPos);
	RotateLightVector(e->angles, e->origin, lightPos, relativelightpos); 
    GL_DrawAliasStencilShadow(paliashdr, pose, relativelightpos);
	glPopMatrix ();
}

/*
=================
GL_StencilShadowing

Go through each frame and draw stencil shadows for appropriate ents
-rww
=================
*/
void GL_StencilShadowing(void)
{
	if (r_shadows.value < 2 || g_numStencilEnts <= 0)
	{
		return;
	}

	//Set us up for drawing our passes onto the stencil buffer
	glPushAttrib( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_ENABLE_BIT | GL_POLYGON_BIT | GL_STENCIL_BUFFER_BIT );
	glDisable (GL_TEXTURE_2D);
	glDepthMask( GL_FALSE );

	if (r_shadows.value == 2)
	{ //> 2 is a debug feature to draw the shadows outside of the stencil buffer
		glEnable( GL_STENCIL_TEST );
		glColorMask( GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE );
	}
	else
	{
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
		glPolygonMode (GL_FRONT_AND_BACK, GL_LINE);
	}

	glStencilFunc( GL_ALWAYS, 1, 0xFFFFFFFF );
	glDepthFunc(GL_LESS);

	while (g_numStencilEnts > 0)
	{ //go through the list and do all the shadows
		g_numStencilEnts--;
		GL_StencilShadowModel(g_stencilEnts[g_numStencilEnts]);
	}

	glCullFace(GL_FRONT);
	glDepthFunc(GL_LEQUAL);

	glColorMask( GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE );

	//draw a full screen rectangle now - wherever we draw
	//is where our "shadows" will actually be seen.
	glColor4f( 0.0f, 0.0f, 0.0f, 0.6f );
	glStencilFunc( GL_NOTEQUAL, 0, 0xFFFFFFFF );

	glStencilOp( GL_KEEP, GL_KEEP, GL_KEEP );

	glEnable (GL_BLEND);
	glPushMatrix(); //save the current view matrix

	glLoadIdentity ();

	glRotatef (-90,  1, 0, 0);
	glRotatef (90,  0, 0, 1);
	glBegin (GL_QUADS);
		glVertex3f (10, 100, 100);
		glVertex3f (10, -100, 100);
		glVertex3f (10, -100, -100);
		glVertex3f (10, 100, -100);
	glEnd ();

	glPopMatrix(); //so we aren't stuck with the identity matrix
	glPopAttrib(); //pop back our previous enable/etc. state so we don't have to bother switching it all back
}

/*
=================
R_SetupAliasFrame

=================
*/
//void R_SetupAliasFrame (int frame, aliashdr_t *paliashdr)
void R_SetupAliasFrame (int frame, aliashdr_t *paliashdr, float alpha)
{
	int				pose, numposes;
	float			interval;

	if ((frame >= paliashdr->numframes) || (frame < 0))
	{
		Con_DPrintf ("R_AliasSetupFrame: no such frame %d\n", frame);
		frame = 0;
	}

	pose = paliashdr->frames[frame].firstpose;
	numposes = paliashdr->frames[frame].numposes;

	if (numposes > 1)
	{
		interval = paliashdr->frames[frame].interval;
		pose += (int)(cl.time / interval) % numposes;
	}

	//GL_DrawAliasFrame (paliashdr, pose);
	GL_DrawAliasFrame (paliashdr, pose, alpha);
}

/*
=================
R_SetupAliasBlendedFrame

fenix@io.com: model animation interpolation
=================
*/
void R_SetupAliasBlendedFrame (int frame, aliashdr_t *paliashdr, entity_t* e)
         {
            int   pose;
            int   numposes;
            float blend;

            if ((frame >= paliashdr->numframes) || (frame < 0))
               {
               Con_DPrintf ("R_AliasSetupFrame: no such frame %d\n", frame);
               frame = 0;
               }

            pose = paliashdr->frames[frame].firstpose;
            numposes = paliashdr->frames[frame].numposes;

            if (numposes > 1)
               {
               e->frame_interval = paliashdr->frames[frame].interval;
               pose += (int)(cl.time / e->frame_interval) % numposes;
               }
            else 
               {
               /* One tenth of a second is a good for most Quake animations.
               If the nextthink is longer then the animation is usually meant to pause
               (e.g. check out the shambler magic animation in shambler.qc).  If its
               shorter then things will still be smoothed partly, and the jumps will be
               less noticable because of the shorter time.  So, this is probably a good
               assumption. */
               e->frame_interval = 0.1;
               }

            if (e->pose2 != pose)
               {
               e->frame_start_time = realtime;
               e->pose1 = e->pose2;
               e->pose2 = pose;
               blend = 0;
               }
            else
               {
               blend = (realtime - e->frame_start_time) / e->frame_interval;
               }
      
            // wierd things start happening if blend passes 1
            if (cl.paused || blend > 1) blend = 1;
            
            GL_DrawAliasBlendedFrame (paliashdr, e->pose1, e->pose2, blend);
         }

/*
=============
GL_DrawAliasBlendedShadow

fenix@io.com: model animation interpolation
=============

void GL_DrawAliasBlendedShadow (aliashdr_t *paliashdr, int pose1, int pose2, entity_t* e)
{
            trivertx_t* verts1;
            trivertx_t* verts2;
            int*        order;
            vec3_t      point1;
            vec3_t      point2;
            vec3_t      d;
            float       height;
            float       lheight;
            int         count;
            float       blend;

            blend = (realtime - e->frame_start_time) / e->frame_interval;

            if (blend > 1) blend = 1;

            lheight = e->origin[2] - lightspot[2];
            height  = -lheight + 1.0;

            verts1 = (trivertx_t *)((byte *)paliashdr + paliashdr->posedata);
            verts2 = verts1;

            verts1 += pose1 * paliashdr->poseverts;
            verts2 += pose2 * paliashdr->poseverts;

            order = (int *)((byte *)paliashdr + paliashdr->commands);

            for (;;)
               {
               // get the vertex count and primitive type
               count = *order++;

               if (!count) break;

               if (count < 0)
                  {
                  count = -count;
                  glBegin (GL_TRIANGLE_FAN);
                  }
               else
                  {
                  glBegin (GL_TRIANGLE_STRIP);
                  }

               do
                  {
                  order += 2;

                  point1[0] = verts1->v[0] * paliashdr->scale[0] + paliashdr->scale_origin[0];
                  point1[1] = verts1->v[1] * paliashdr->scale[1] + paliashdr->scale_origin[1];
                  point1[2] = verts1->v[2] * paliashdr->scale[2] + paliashdr->scale_origin[2];
        
                  point1[0] -= shadevector[0]*(point1[2]+lheight);
                  point1[1] -= shadevector[1]*(point1[2]+lheight);

                  point2[0] = verts2->v[0] * paliashdr->scale[0] + paliashdr->scale_origin[0];
                  point2[1] = verts2->v[1] * paliashdr->scale[1] + paliashdr->scale_origin[1];
                  point2[2] = verts2->v[2] * paliashdr->scale[2] + paliashdr->scale_origin[2];

                  point2[0] -= shadevector[0]*(point2[2]+lheight);
                  point2[1] -= shadevector[1]*(point2[2]+lheight);

                  VectorSubtract(point2, point1, d);

                  glVertex3f (point1[0] + (blend * d[0]), 
                              point1[1] + (blend * d[1]), 
                              height);

                  verts1++;
                  verts2++;
                  } while (--count);

                  glEnd ();
               }       
}
*/
void GL_DrawAliasBlendedShadow (aliashdr_t *paliashdr, int pose1, int pose2, entity_t* e)
{
            trivertx_t* verts1;
            trivertx_t* verts2;
            int*        order;
            vec3_t      point1;
            vec3_t      point2;
            vec3_t      d;
            float       height;
            float       lheight;
            int         count;
            float       blend;

	// muff@yakko.globalnet.co.uk - 17 Mar 2000
	trace_t		steptrace, downtrace;
	vec3_t		dest,stop,downmove;
	float		s1,c1; //this will store the sin and cos values as they are used so heavily
	//end of muff


            blend = (realtime - e->frame_start_time) / e->frame_interval;

            if (blend > 1) blend = 1;

            lheight = e->origin[2] - lightspot[2];
            height  = -lheight;// + 1.0; //was 1.0 - muff


	// muff@yakko.globalnet.co.uk - 17 Mar 2000
	// better shadowing, now takes angle of ground into account	
	// cast a traceline into the floor directly below the player
	// and gets normals from this
	// for proper surface shadowing, each point should be taken
	// seperately, we do it this way for speed
	// NOTE need fixing for when at liquid surface
	VectorCopy (currententity->origin, downmove);
	downmove[2] = downmove[2] - 4096;
	memset (&downtrace, 0, sizeof(downtrace));
	SV_RecursiveHullCheck (cl.worldmodel->hulls, 0, 0, 1, currententity->origin, downmove, &downtrace);

	// calculate the all important angles
	s1 = sin( currententity->angles[1]/180*M_PI);
	c1 = cos( currententity->angles[1]/180*M_PI);

	// end of muff

            verts1 = (trivertx_t *)((byte *)paliashdr + paliashdr->posedata);
            verts2 = verts1;

            verts1 += pose1 * paliashdr->poseverts;
            verts2 += pose2 * paliashdr->poseverts;

            order = (int *)((byte *)paliashdr + paliashdr->commands);

            for (;;)
               {
               // get the vertex count and primitive type
               count = *order++;

               if (!count) break;

               if (count < 0)
                  {
                  count = -count;
                  glBegin (GL_TRIANGLE_FAN);
                  }
               else
                  {
                  glBegin (GL_TRIANGLE_STRIP);
                  }

               do
                  {
                  order += 2;

                  point1[0] = verts1->v[0] * paliashdr->scale[0] + paliashdr->scale_origin[0];
                  point1[1] = verts1->v[1] * paliashdr->scale[1] + paliashdr->scale_origin[1];
                  point1[2] = verts1->v[2] * paliashdr->scale[2] + paliashdr->scale_origin[2];
        
                  point1[0] -= shadevector[0]*(point1[2]+lheight);
                  point1[1] -= shadevector[1]*(point1[2]+lheight);

                  point2[0] = verts2->v[0] * paliashdr->scale[0] + paliashdr->scale_origin[0];
                  point2[1] = verts2->v[1] * paliashdr->scale[1] + paliashdr->scale_origin[1];
                  point2[2] = verts2->v[2] * paliashdr->scale[2] + paliashdr->scale_origin[2];

                  point2[0] -= shadevector[0]*(point2[2]+lheight); 
                  point2[1] -= shadevector[1]*(point2[2]+lheight); 

                  VectorSubtract(point2, point1, d);


			// muff@yakko.globalnet.co.uk - 17 Mar 2000
			point1[0] = point1[0] + (blend * d[0]);  //+lheight); //muff
			point1[1] = point1[1] + (blend * d[1]);  //+lheight); //muff
			point1[2] = point1[2] + (blend * d[2]);  //+lheight); //muff

			//drop it down to floor
			point1[2] =  - (currententity->origin[2] - downtrace.endpos[2]) ;

			//now move the z-coordinate as appropriate
			point1[2] += (	(point1[1] * (s1 * downtrace.plane.normal[0])) -
				        (point1[0] * (c1 * downtrace.plane.normal[0])) -
					(point1[0] * (s1 * downtrace.plane.normal[1])) - 
					(point1[1] * (c1 * downtrace.plane.normal[1])) 
					) +  ((1.0 - downtrace.plane.normal[2])*20)
					+ 0.2 ;


			// now load the point we have just calculated
			glVertex3fv (point1);

			// remove this as we have new commands in above
			//  glVertex3f (point1[0] + (blend * d[0]), 
                        //  point1[1] + (blend * d[1]), 
                        //  height);
	                

	            //end of muff

				  verts1++;
                  verts2++;
                  } while (--count);

                  glEnd ();
               }       
}

/*
=================
R_DrawAliasModel
=================

void R_DrawAliasModel (entity_t *e)
{
	int			i;
	int			lnum;
	vec3_t		dist;
	float		add;
	model_t		*clmodel;
	vec3_t		mins, maxs;
	aliashdr_t	*paliashdr;
	float		an;
	int			anim;

	glDisable(GL_ALPHA_TEST);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	clmodel = currententity->model;

	VectorAdd (currententity->origin, clmodel->mins, mins);
	VectorAdd (currententity->origin, clmodel->maxs, maxs);

	if (R_CullBox (mins, maxs))
		return;


	VectorCopy (currententity->origin, r_entorigin);
	VectorSubtract (r_origin, r_entorigin, modelorg);

	//
	// get lighting information
	//

	ambientlight = shadelight = R_LightPoint (currententity->origin);

	// allways give the gun some light
	if (e == &cl.viewent && ambientlight < 24)
		ambientlight = shadelight = 24;

	for (lnum=0 ; lnum<MAX_DLIGHTS ; lnum++)
	{
		if (cl_dlights[lnum].die >= cl.time)
		{
			VectorSubtract (currententity->origin,
							cl_dlights[lnum].origin,
							dist);
			add = cl_dlights[lnum].radius - Length(dist);

			if (add > 0) {
				ambientlight += add;
				//ZOID models should be affected by dlights as well
				shadelight += add;
			}
		}
	}

	// clamp lighting so it doesn't overbright as much
	if (ambientlight > 128)
		ambientlight = 128;
	if (ambientlight + shadelight > 192)
		shadelight = 192 - ambientlight;

	// ZOID: never allow players to go totally black
	i = currententity - cl_entities;
	if (i >= 1 && i<=cl.maxclients // && !strcmp (currententity->model->name, "progs/player.mdl") un-// )
		if (ambientlight < 8)
			ambientlight = shadelight = 8;

	// HACK HACK HACK -- no fullbright colors, so make torches full light
	if (!strcmp (clmodel->name, "progs/flame2.mdl")
		|| !strcmp (clmodel->name, "progs/flame.mdl") )
		ambientlight = shadelight = 256;

	shadedots = r_avertexnormal_dots[((int)(e->angles[1] * (SHADEDOT_QUANT / 360.0))) & (SHADEDOT_QUANT - 1)];
	shadelight = shadelight / 200.0;
	
	an = e->angles[1]/180*M_PI;
	shadevector[0] = cos(-an);
	shadevector[1] = sin(-an);
	shadevector[2] = 1;
	VectorNormalize (shadevector);

	//
	// locate the proper data
	//
	paliashdr = (aliashdr_t *)Mod_Extradata (currententity->model);

	c_alias_polys += paliashdr->numtris;

	//
	// draw all the triangles
	//

	GL_DisableMultitexture();

    glPushMatrix ();
	R_RotateForEntity (e);

	if (!strcmp (clmodel->name, "progs/eyes.mdl") && gl_doubleeyes.value) {
		glTranslatef (paliashdr->scale_origin[0], paliashdr->scale_origin[1], paliashdr->scale_origin[2] - (22 + 8));
// double size of eyes, since they are really hard to see in gl
		glScalef (paliashdr->scale[0]*2, paliashdr->scale[1]*2, paliashdr->scale[2]*2);
	} else {
		glTranslatef (paliashdr->scale_origin[0], paliashdr->scale_origin[1], paliashdr->scale_origin[2]);
		glScalef (paliashdr->scale[0], paliashdr->scale[1], paliashdr->scale[2]);
	}

	anim = (int)(cl.time*10) & 3;
    GL_Bind(paliashdr->gl_texturenum[currententity->skinnum][anim]);

	// we can't dynamically colormap textures, so they are cached
	// seperately for the players.  Heads are just uncolored.
	if (currententity->colormap != vid.colormap && !gl_nocolors.value)
	{
		i = currententity - cl_entities;
		if (i >= 1 && i<=cl.maxclients // && !strcmp (currententity->model->name, "progs/player.mdl") un-//)
		    GL_Bind(playertextures - 1 + i);
	}

	if (gl_smoothmodels.value)
		glShadeModel (GL_SMOOTH);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	if (gl_affinemodels.value)
		glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);

	R_SetupAliasFrame (currententity->frame, paliashdr);

	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	glShadeModel (GL_FLAT);
	if (gl_affinemodels.value)
		glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

	glPopMatrix ();

	if (r_shadows.value)
	{
		glPushMatrix ();
		R_RotateForEntity (e);
		glDisable (GL_TEXTURE_2D);
		glEnable (GL_BLEND);
		glColor4f (0,0,0,0.5);
		GL_DrawAliasShadow (paliashdr, lastposenum);
		glEnable (GL_TEXTURE_2D);
		glDisable (GL_BLEND);
		glColor4f (1,1,1,1);
		glPopMatrix ();
	}

		glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		glDisable(GL_BLEND);
		glDisable(GL_ALPHA_TEST);
}
*/
/* 
================= 
R_DrawAliasModel 
================= 
*/ 
void R_DrawAliasModel (entity_t *e) 
{ 
   int	 maxnumsurf;
   int   i, j; 
   int   lnum; 
   vec3_t  dist; 
   float  add; 
   model_t  *clmodel; 
   vec3_t  mins, maxs; 
   aliashdr_t *paliashdr; 
   trivertx_t *verts, *v; 
   int   index; 
   float  s, t, an; 
   int   anim; 

   //muff@yakko.globalnet.co.uk - 05 Mar 00 - added
	trace_t		steptrace, downtrace;
	vec3_t		dest,stop,downmove;
	//end of muff

   clmodel = currententity->model; 
   VectorAdd (currententity->origin, clmodel->mins, mins); 
   VectorAdd (currententity->origin, clmodel->maxs, maxs); 
   
	//Since we are using stencil shadows and they tend to stretch
	//outside of the viewable area of the entity itself, we want
	//to draw them regardless of whether we pass the cull check.
	if (r_shadows.value >= 2)
	{ //Add this guy to our list.
		if (g_numStencilEnts < MAX_STENCIL_ENTS)
		{
			g_stencilEnts[g_numStencilEnts] = currententity;
			g_numStencilEnts++;
		}
	}
   
   if (e != &cl.viewent)
   {
	   if (R_CullBox (mins,maxs))
	   {
		   return;
	   }
   }
/*	
   if (R_CullBox (mins, maxs)) 
      return; 
*/

/*
   if (!strcmp (clmodel->name, "progs/flame.mdl"))
		AddFire(e->origin, 4);

   if (!strcmp (clmodel->name, "progs/flame2.mdl"))
   {
		AddFire(e->origin, 10);
		return; //do not draw the big fire model, its just a place holder for the particles
   }
*/
   VectorCopy (currententity->origin, r_entorigin); 
   VectorSubtract (r_origin, r_entorigin, modelorg); 

   // get lighting information
   //ambientlight = shadelight = R_LightPoint (currententity->origin); // old code
   R_LightPoint(currententity->origin); // LordHavoc: lightcolor is all that matters from this

	#ifdef Q3MODELS 
	if (clmodel->aliastype == MD3IDHEADER)
	{ 
      //do nothing for testing 
      R_DrawQ3Model (e); 
	}
	else 
	#endif 


   // allways give the gun some light 
   // LH start
//   if (e == &cl.viewent && ambientlight < 24) 
//      ambientlight = shadelight = 24; 
	if (e == &cl.viewent)
	{
		if (lightcolor[0] < 24)
			lightcolor[0] = 24;
		if (lightcolor[1] < 24)
			lightcolor[1] = 24;
		if (lightcolor[2] < 24)
			lightcolor[2] = 24;
	}
	// LH end

   for (lnum=0 ; lnum < MAX_DLIGHTS ; lnum++)
   { 
      if (cl_dlights[lnum].die >= cl.time) 
      { 
         VectorSubtract (currententity->origin, 
            cl_dlights[lnum].origin, 
            dist); 
         add = cl_dlights[lnum].radius - Length(dist); 
		 /*
         if (add > 0)
            {
            ambientlight += add; 
            shadelight += add; 
            } 
         }*/
		 if (add > 0)
		 {
		 	lightcolor[0] += add * cl_dlights[lnum].color[0];
		 	lightcolor[1] += add * cl_dlights[lnum].color[1];
		 	lightcolor[2] += add * cl_dlights[lnum].color[2];
		 }
      } 
	}
   // clamp lighting so it doesn't overbright as much 
	  /*
   if (ambientlight > 128) 
      ambientlight = 128; 
   if (ambientlight + shadelight > 192) 
      shadelight = 192 - ambientlight; 
*/
   i = currententity - cl_entities; 
   if (i >= 1 && i<=cl.maxclients /* && !strcmp (currententity->model->name, "progs/player.mdl") */) 
   {
		if (lightcolor[0] < 8)
			lightcolor[0] = 8;
		if (lightcolor[1] < 8)
			lightcolor[1] = 8;
		if (lightcolor[2] < 8)
			lightcolor[2] = 8;
	}
//      if (ambientlight < 8) 
//         ambientlight = shadelight = 8; 

   // HACK HACK HACK -- no fullbright colors, so make torches full light
   if (!strcmp (clmodel->name, "progs/flame2.mdl") 
      || !strcmp (clmodel->name, "progs/flame.mdl")) 
      //ambientlight = shadelight = 256; 
		lightcolor[0] = lightcolor[1] = lightcolor[2] = 256;

   shadedots = r_avertexnormal_dots[((int)(e->angles[1] * (SHADEDOT_QUANT / 360.0))) & (SHADEDOT_QUANT - 1)]; 
   //shadelight = shadelight / 200.0; 
	VectorScale(lightcolor, 1.0f / 200.0f, lightcolor);

   an = e->angles[1]/180*M_PI; 
   shadevector[0] = cos(-an); 
   shadevector[1] = sin(-an); 
   shadevector[2] = 1; 
   VectorNormalize (shadevector); 

   // locate the proper data 
   paliashdr = (aliashdr_t *)Mod_Extradata (currententity->model); 
   c_alias_polys += paliashdr->numtris; 

   // draw all the triangles
   /*
   GL_DisableMultitexture(); 
   glPushMatrix (); 
   //R_RotateForEntity (e); 
   // fenix@io.com: model transform interpolation
   if (r_interpolate_model_transform.value)
      {
      R_BlendedRotateForEntity (e);
      }
   else
      {
      R_RotateForEntity (e);
      }
	*/
    GL_DisableMultitexture();


    glPushMatrix ();

         // fenix@io.com: model transform interpolation
         //if (r_interpolate_model_transform.value)
		 if (r_interpolate_model_t.value)
            {
               R_BlendedRotateForEntity (e, 0);
      		}
         else
            {
               R_RotateForEntity (e);
            }

   if (!strcmp (clmodel->name, "progs/eyes.mdl") && gl_doubleeyes.value)
      {
      glTranslatef (paliashdr->scale_origin[0], paliashdr->scale_origin[1], paliashdr->scale_origin[2] - (22 + 8)); 
      // double size of eyes, since they are really hard to see in gl 
      glScalef (paliashdr->scale[0]*2, paliashdr->scale[1]*2, paliashdr->scale[2]*2); 
      }
   else
      {
      glTranslatef (paliashdr->scale_origin[0], paliashdr->scale_origin[1], paliashdr->scale_origin[2]); 
      glScalef (paliashdr->scale[0], paliashdr->scale[1], paliashdr->scale[2]); 
      } 
	
   anim = (int)(cl.time*10) & 3; 
   GL_Bind(paliashdr->gl_texturenum[currententity->skinnum][anim]); 

   // we can't dynamically colormap textures, so they are cached 
   // seperately for the players.  Heads are just uncolored. 
   if (currententity->colormap != vid.colormap && !gl_nocolors.value) 
      { 
      i = currententity - cl_entities; 
      if (i >= 1 && i<=cl.maxclients /* && !strcmp (currententity->model->name, "progs/player.mdl") */) 
      GL_Bind(playertextures - 1 + i); 
      } 

   if (gl_smoothmodels.value) 
      glShadeModel (GL_SMOOTH); 
   glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE); 

   if (gl_affinemodels.value) 
      glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST); 
   
   // Topaz: Interpolation code
   
   if (r_interpolations.value && currententity->interpolation < r_interpolations.value) 
      currententity->interpolation++;  
   if (r_interpolations.value) 
   {  
   if (gl_asynch.value) 
      { 
      if (currententity->interpolation >= r_interpolations.value) 
         { 
         currententity->last_frame = currententity->current_frame; 
         currententity->current_frame = currententity->frame; 
         currententity->interpolation = 1; 
         } 
      } 
   else 
      { 
      if (currententity->frame != currententity->current_frame) 
         { 
         currententity->last_frame = currententity->current_frame; 
         currententity->current_frame = currententity->frame; 
         currententity->interpolation = 1; 
         } 
      } 
   R_SetupAliasInterpolatedFrame (currententity->current_frame,  
                                  currententity->last_frame,  
                                  currententity->interpolation, 
                                  paliashdr, e->alpha); 
   } 
else 
   //R_SetupAliasFrame (currententity->frame, paliashdr); 

         // fenix@io.com: model animation interpolation
         //if (r_interpolate_model_animation.value)
		 if (r_interpolate_model_a.value)
            {
            R_SetupAliasBlendedFrame (currententity->frame, paliashdr, currententity);
            }
         else
            {
            //R_SetupAliasFrame (currententity->frame, paliashdr);
			R_SetupAliasFrame (currententity->frame, paliashdr, e->alpha);
            }


   glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE); 

   glShadeModel (GL_FLAT); 
   if (gl_affinemodels.value) 
   glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST); 

   glPopMatrix (); 
		/*
   if (r_shadows.value) 
      { 
      glPushMatrix (); 
      //R_RotateForEntity (e);
	  
      // fenix@io.com: model transform interpolation
      if (r_interpolate_model_transform.value)
         {
         R_BlendedRotateForEntity (e);
         }
      else
         {
         R_RotateForEntity (e);
         }

      glDisable (GL_TEXTURE_2D); 
      glEnable (GL_BLEND); 
      glColor4f (0,0,0,0.5); 
      GL_DrawAliasShadow (paliashdr, lastposenum); 
      // fenix@io.com: model animation interpolation
      if (r_interpolate_model_animation.value)
         {
         GL_DrawAliasBlendedShadow (paliashdr, lastposenum0, lastposenum, currententity);
         }
      else
         {
         GL_DrawAliasShadow (paliashdr, lastposenum);
         }

      glEnable (GL_TEXTURE_2D); 
      glDisable (GL_BLEND); 
      glColor4f (1,1,1,1); 
      glPopMatrix (); 
      }
	  */
	//if (r_shadows.value)
    if (r_shadows.value && r_shadows.value < 2)
	{


		// muff@yakko.globalnet.co.uk - 05 Mar 00 
		// lightning bolts really shouldn't cast a shadow
		// Neither should rockets
		// lots of other to go in here no doubt, but we'll add them as and when
		if ( (!strncmp (clmodel->name, "progs/bolt",10)) ||
		     (!strcmp  (clmodel->name, "progs/missile.mdl")) ||
		     (!strcmp  (clmodel->name, "progs/laser.mdl")) ||
		     (!strcmp  (clmodel->name, "progs/lavaball.mdl")) ||
		     (!strcmp  (clmodel->name, "progs/quaddama.mdl")) ||
		     (!strcmp  (clmodel->name, "progs/invulner.mdl")) ||
			 (!strcmp  (clmodel->name, "progs/emp_sphere.mdl")) ||
			 (!strcmp  (clmodel->name, "progs/flame2.mdl"))
		   )
		   return;

		// in first person the engine currently draws the player's weapon,
		// but no player. This means a weapon shadow is drawn
		// we'll remove weapon shadow until this is fixed
		if (!strncmp (clmodel->name, "progs/v_",8))
			return;
		// end of muff

		glPushMatrix ();

		// muff@yakko.globalnet.co.uk - 05 Mar 00 
		// this determines the shadow vector relative to origin of entity
		// changed always be directly below entity for the moment.
		// will alter later (hopefully) when I got proper light source shadows working
		shadevector[0] = 0;
		shadevector[1] = 0;
		shadevector[2] = 1;
		VectorNormalize (shadevector);

		//original code
		//		R_RotateForEntity (e);

		//muff@yakko.globalnet.co.uk - 050300 - added
		// this translates the location of the shadow, and rotates in horizontal plane only
		// this means that shadows now work properly for chase_cam, grenade etc
		// hence removal of code above
		//if (r_interpolate_model_animation.value)
		if (r_interpolate_model_a.value)
		 R_BlendedRotateForEntity (e, 1); // we've re-written this to work properly		 
		else
		{
		 glTranslatef (e->origin[0],  e->origin[1],  e->origin[2]);
	 	 glRotatef (e->angles[1],  0, 0, 1);
		}
		//end of muff
		// muff@yakko.globalnet.co.uk - 14 Mar 2000
		// this is so we can calculate height of the ground accurately
		VectorCopy (e->origin, downmove);

		downmove[2] = downmove[2] - 4096;
		memset (&downtrace, 0, sizeof(downtrace));
		SV_RecursiveHullCheck (cl.worldmodel->hulls, 0, 0, 1, e->origin, downmove, &downtrace);
		// end of muff

		glDisable (GL_TEXTURE_2D);
		glEnable (GL_BLEND);
		
		// original code
		//	glColor4f (0,0,0,0.5);
		

		// muff@yakko.globalnet.co.uk - 14 Mar 2000
		// set the shade of the shadow based on height off ground
		glColor4f (0,0,0,1.0 - ((mins[2]-downtrace.endpos[2])/60)); 
		//end of muff

		//original code
		//	GL_DrawAliasShadow (paliashdr, lastposenum);

		// fenix@io.com: model animation interpolation
		//if (r_interpolate_model_animation.value)
		if (r_interpolate_model_a.value)
		 GL_DrawAliasBlendedShadow (paliashdr, lastposenum0, lastposenum, currententity);
		else
		 GL_DrawAliasShadow (paliashdr, lastposenum);

		glEnable (GL_TEXTURE_2D);
		glDisable (GL_BLEND);
		glColor4f (1,1,1,1);
		glPopMatrix ();
	}
	
}

//==================================================================================

/*
=============
R_DrawEntitiesOnList
=============
*/
void R_DrawEntitiesOnList (void)
{
	int		i;

	if (!r_drawentities.value)
	{
		return;
	}
	// draw sprites seperately, because of alpha blending
	for (i=0 ; i<cl_numvisedicts ; i++)
	{
		currententity = cl_visedicts[i];

		if (currententity == &cl_entities[cl.viewentity]) 
			currententity->angles[0] *= 0.3;

		switch (currententity->model->type)
		{
		case mod_alias:
			R_DrawAliasModel (currententity);
			break;

		case mod_brush:
			R_DrawBrushModel (currententity);
			break;

		default:
			break;
		}
	}

	for (i=0 ; i<cl_numvisedicts ; i++)
	{
		currententity = cl_visedicts[i];

		switch (currententity->model->type)
		{
		case mod_sprite:
			R_DrawSpriteModel (currententity);
			break;
		}
	}
}

/*
=============
R_DrawViewModel
=============
*/
void R_DrawViewModel (void)
{
	/*float		ambient[4], diffuse[4];
	int			j;
	int			lnum;
	vec3_t		dist;
	float		add;
	dlight_t	*dl;
	int			ambientlight, shadelight;*/

	if (!r_drawviewmodel.value)
		return;

	if (chase_active.value)
		return;

	if (envmap)
		return;

	if (!r_drawentities.value)
		return;

	if (cl.items & IT_INVISIBILITY)
		return;

	if (cl.stats[STAT_HEALTH] <= 0)
		return;

	currententity = &cl.viewent;
	if (!currententity->model)
		return;
/*
	j = R_LightPoint (currententity->origin);

	if (j < 24)
		j = 24;		// allways give some light on gun
	ambientlight = j;
	shadelight = j;

// add dynamic lights		
	for (lnum=0 ; lnum<MAX_DLIGHTS ; lnum++)
	{
		dl = &cl_dlights[lnum];
		if (!dl->radius)
			continue;
		if (!dl->radius)
			continue;
		if (dl->die < cl.time)
			continue;

		VectorSubtract (currententity->origin, dl->origin, dist);
		add = dl->radius - Length(dist);
		if (add > 0)
			ambientlight += add;
	}

	ambient[0] = ambient[1] = ambient[2] = ambient[3] = (float)ambientlight / 128;
	diffuse[0] = diffuse[1] = diffuse[2] = diffuse[3] = (float)shadelight / 128;
*/
	// hack the depth range to prevent view model from poking into walls
	glDepthRange (gldepthmin, gldepthmin + 0.3*(gldepthmax-gldepthmin));
	//R_DrawAliasModel (currententity);
    // fenix@io.com: model transform interpolation
    //old_interpolate_model_transform = r_interpolate_model_transform.value;
	old_interpolate_model_transform = r_interpolate_model_t.value;
    //r_interpolate_model_transform.value = false;
	r_interpolate_model_t.value = false;
    R_DrawAliasModel (currententity);
    //r_interpolate_model_transform.value = old_interpolate_model_transform;
	r_interpolate_model_t.value = old_interpolate_model_transform;

	glDepthRange (gldepthmin, gldepthmax);
}


/*
============
R_PolyBlend
============

void R_PolyBlend (void)
{
	if (!gl_polyblend.value)
		return;
	if (!v_blend[3])
		return;

	GL_DisableMultitexture();

	glDisable (GL_ALPHA_TEST);
	glEnable (GL_BLEND);
	glDisable (GL_DEPTH_TEST);
	glDisable (GL_TEXTURE_2D);

    glLoadIdentity ();

    glRotatef (-90,  1, 0, 0);	    // put Z going up
    glRotatef (90,  0, 0, 1);	    // put Z going up

	glColor4fv (v_blend);

	glBegin (GL_QUADS);

	glVertex3f (10, 100, 100);
	glVertex3f (10, -100, 100);
	glVertex3f (10, -100, -100);
	glVertex3f (10, 100, -100);
	glEnd ();

	glDisable (GL_BLEND);
	glEnable (GL_TEXTURE_2D);
	glEnable (GL_ALPHA_TEST);
}
*/
//some code from LordHavoc early DP version

void R_PolyBlend (void)
{
	if (!gl_polyblend.value)
		return;

	GL_DisableMultitexture();

	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glEnable (GL_BLEND);
	glDisable (GL_DEPTH_TEST);
	glDisable (GL_TEXTURE_2D);

      	glLoadIdentity ();

      	glRotatef (-90,  1, 0, 0);	    // put Z going up
      	glRotatef (90,  0, 0, 1);	    // put Z going up

	if (v_blend[3])
	{
		glColor4fv (v_blend);

		glBegin (GL_QUADS);

		glVertex3f (10, 100, 100);
		glVertex3f (10, -100, 100);
		glVertex3f (10, -100, -100);
		glVertex3f (10, 100, -100);
		glEnd ();

	}

	//gamma trick based on LordHavoc - muff
	if (v_gamma.value != 1)
		DoGamma(); 

	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable (GL_BLEND); // muff
	glEnable (GL_TEXTURE_2D);
	glEnable (GL_ALPHA_TEST); //muff
}

int SignbitsForPlane (mplane_t *out)
{
	int	bits, j;

	// for fast box on planeside test

	bits = 0;
	for (j=0 ; j<3 ; j++)
	{
		if (out->normal[j] < 0)
			bits |= 1<<j;
	}
	return bits;
}


void R_SetFrustum (void)
{
	int		i;

	if (r_refdef.fov_x == 90) 
	{
		// front side is visible

		VectorAdd (vpn, vright, frustum[0].normal);
		VectorSubtract (vpn, vright, frustum[1].normal);

		VectorAdd (vpn, vup, frustum[2].normal);
		VectorSubtract (vpn, vup, frustum[3].normal);
	}
	else
	{
		// rotate VPN right by FOV_X/2 degrees
		RotatePointAroundVector( frustum[0].normal, vup, vpn, -(90-r_refdef.fov_x / 2 ) );
		// rotate VPN left by FOV_X/2 degrees
		RotatePointAroundVector( frustum[1].normal, vup, vpn, 90-r_refdef.fov_x / 2 );
		// rotate VPN up by FOV_X/2 degrees
		RotatePointAroundVector( frustum[2].normal, vright, vpn, 90-r_refdef.fov_y / 2 );
		// rotate VPN down by FOV_X/2 degrees
		RotatePointAroundVector( frustum[3].normal, vright, vpn, -( 90 - r_refdef.fov_y / 2 ) );
	}

	for (i=0 ; i<4 ; i++)
	{
		frustum[i].type = PLANE_ANYZ;
		frustum[i].dist = DotProduct (r_origin, frustum[i].normal);
		frustum[i].signbits = SignbitsForPlane (&frustum[i]);
	}
}



/*
===============
R_SetupFrame
===============
*/
void R_SetupFrame (void)
{

// don't allow cheats in multiplayer
	if (cl.maxclients > 1)
		Cvar_Set ("r_fullbright", "0");

	R_AnimateLight ();

	r_framecount++;

// build the transformation matrix for the given view angles
	VectorCopy (r_refdef.vieworg, r_origin);

	AngleVectors (r_refdef.viewangles, vpn, vright, vup);

// current viewleaf
	r_oldviewleaf = r_viewleaf;
	r_viewleaf = Mod_PointInLeaf (r_origin, cl.worldmodel);

	V_SetContentsColor (r_viewleaf->contents);
	V_CalcBlend ();

	r_cache_thrash = false;

	c_brush_polys = 0;
	c_alias_polys = 0;

}


void MYgluPerspective( GLdouble fovy, GLdouble aspect,
		     GLdouble zNear, GLdouble zFar )
{
   GLdouble xmin, xmax, ymin, ymax;

   ymax = zNear * tan( fovy * M_PI / 360.0 );
   ymin = -ymax;

   xmin = ymin * aspect;
   xmax = ymax * aspect;

   glFrustum( xmin, xmax, ymin, ymax, zNear, zFar );
}


/*
=============
R_SetupGL
=============
*/
void R_SetupGL (void)
{
	float	screenaspect;
	extern	int glwidth, glheight;
	int		x, x2, y2, y, w, h;

	//
	// set up viewpoint
	//
	glMatrixMode(GL_PROJECTION);
    glLoadIdentity ();
	x = r_refdef.vrect.x * glwidth/vid.width;
	x2 = (r_refdef.vrect.x + r_refdef.vrect.width) * glwidth/vid.width;
	y = (vid.height-r_refdef.vrect.y) * glheight/vid.height;
	y2 = (vid.height - (r_refdef.vrect.y + r_refdef.vrect.height)) * glheight/vid.height;

	// fudge around because of frac screen scale
	if (x > 0)
		x--;
	if (x2 < glwidth)
		x2++;
	if (y2 < 0)
		y2--;
	if (y < glheight)
		y++;

	w = x2 - x;
	h = y - y2;

	if (envmap)
	{
		x = y2 = 0;
		w = h = 256;
	}

	glViewport (glx + x, gly + y2, w, h);
    screenaspect = (float)r_refdef.vrect.width/r_refdef.vrect.height;
//	yfov = 2*atan((float)r_refdef.vrect.height/r_refdef.vrect.width)*180/M_PI;
    MYgluPerspective (r_refdef.fov_y,  screenaspect,  4,  4096);

	if (mirror)
	{
		if (mirror_plane->normal[2])
			glScalef (1, -1, 1);
		else
			glScalef (-1, 1, 1);
		glCullFace(GL_BACK);
	}
	else
		glCullFace(GL_FRONT);

	glMatrixMode(GL_MODELVIEW);
    glLoadIdentity ();

    glRotatef (-90,  1, 0, 0);	    // put Z going up
    glRotatef (90,  0, 0, 1);	    // put Z going up
    glRotatef (-r_refdef.viewangles[2],  1, 0, 0);
    glRotatef (-r_refdef.viewangles[0],  0, 1, 0);
    glRotatef (-r_refdef.viewangles[1],  0, 0, 1);
    glTranslatef (-r_refdef.vieworg[0],  -r_refdef.vieworg[1],  -r_refdef.vieworg[2]);

	glGetFloatv (GL_MODELVIEW_MATRIX, r_world_matrix);

	//
	// set drawing parms
	//
	if (gl_cull.value)
		glEnable(GL_CULL_FACE);
	else
		glDisable(GL_CULL_FACE);

	glDisable(GL_BLEND);
	glDisable(GL_ALPHA_TEST);
	glEnable(GL_DEPTH_TEST);
}
/*
void *VolumeVertsPointer;

=============
R_ConstructShadowVolume

	Calculate the shadow volume commands for the light.
	(only if dynamic)
=============
/
void R_ConstructShadowVolume(shadowlight_t *light) {

	if (!light->isStatic) {
		PrecalcVolumesForLight(cl.worldmodel);
		light->volumeCmds = &volumeCmdsBuff[0];
		light->volumeVerts = &volumeVertsBuff[0];
		light->lightCmds = &lightCmdsBuff[0];
	}

	VolumeVertsPointer = light->volumeVerts;

	if (gl_var) {
		if (light->numVolumeVerts < AGP_BUFFER_SIZE/4) {
			memcpy(AGP_Buffer,light->volumeVerts,light->numVolumeVerts*4);
			VolumeVertsPointer = AGP_Buffer;
		}
	}
}

/*
=============
R_DrawShadowVolume

Draws the shadow volume, for statics this is the precalc one
for dynamics this is the R_ConstructShadowVolume one.

=============
/
void R_DrawShadowVolume(shadowlight_t *light) {


	glDisable(GL_TEXTURE_2D);
	
/*	if (!light->isStatic) {
		glColor3f(0,1,0);
		DrawVolumeFromCmds (&volumeCmdsBuff[0], &lightCmdsBuff[0]);
	} else {
/
		glColor3f(1,0,0);
		DrawVolumeFromCmds (light->volumeCmds, light->lightCmds, VolumeVertsPointer);
//	}

	glColor3f(1,1,1);
	glEnable(GL_TEXTURE_2D);
}
*/
/*
================
R_RenderScene

r_refdef must be set before the first call
================
*/
void R_RenderScene (void)
{
	//shadowlight_t *l = NULL;

	R_SetupFrame ();

	R_SetFrustum ();

	R_SetupGL ();

	R_MarkLeaves ();	// done here so we know if we're in water

	R_DrawWorld ();		// adds static entities to the list

	S_ExtraUpdate ();	// don't let sound get messed up if going slow

	R_DrawEntitiesOnList ();

	GL_DisableMultitexture();

	GL_StencilShadowing();

	R_RenderDlights ();

	R_DrawParticles ();
	
	//if (sh_worldshadows.value) R_DrawShadowVolume(l);

	//R_DrawDecals();
	//glFogfv(GL_FOG_COLOR, fog_color);
}


/*
=============
R_Clear
=============
*/
void R_Clear (void)
{
	if (r_mirroralpha.value != 1.0)
	{
		if (gl_clear.value)
			glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		else
			glClear (GL_DEPTH_BUFFER_BIT);
		gldepthmin = 0;
		gldepthmax = 0.5;
		glDepthFunc (GL_LEQUAL);
	}
	else if (gl_ztrick.value)
	{
		static int trickframe;

		if (gl_clear.value)
			glClear (GL_COLOR_BUFFER_BIT);

		trickframe++;
		if (trickframe & 1)
		{
			gldepthmin = 0;
			gldepthmax = 0.49999f;
			glDepthFunc (GL_LEQUAL);
		}
		else
		{
			gldepthmin = 1;
			gldepthmax = 0.5;
			glDepthFunc (GL_GEQUAL);
		}
	}
	else
	{
		if (gl_clear.value)
			glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		else
			glClear (GL_DEPTH_BUFFER_BIT);
		gldepthmin = 0;
		gldepthmax = 1;
		glDepthFunc (GL_LEQUAL);
	}

	glDepthRange (gldepthmin, gldepthmax);
	glClear(GL_STENCIL_BUFFER_BIT);
}

/*
=============
R_Mirror
=============
*/
void R_Mirror (void)
{
	float		d;
	msurface_t	*s;
	entity_t	*ent;

	if (!mirror)
		return;

	memcpy (r_base_world_matrix, r_world_matrix, sizeof(r_base_world_matrix));

	d = DotProduct (r_refdef.vieworg, mirror_plane->normal) - mirror_plane->dist;
	VectorMA (r_refdef.vieworg, -2*d, mirror_plane->normal, r_refdef.vieworg);

	d = DotProduct (vpn, mirror_plane->normal);
	VectorMA (vpn, -2*d, mirror_plane->normal, vpn);

	r_refdef.viewangles[0] = -asin (vpn[2])/M_PI*180;
	r_refdef.viewangles[1] = atan2 (vpn[1], vpn[0])/M_PI*180;
	r_refdef.viewangles[2] = -r_refdef.viewangles[2];

	ent = &cl_entities[cl.viewentity];
	if (cl_numvisedicts < MAX_VISEDICTS)
	{
		cl_visedicts[cl_numvisedicts] = ent;
		cl_numvisedicts++;
	}

	gldepthmin = 0.5;
	gldepthmax = 1;
	glDepthRange (gldepthmin, gldepthmax);
	glDepthFunc (GL_LEQUAL);

	R_RenderScene ();
	R_DrawWaterSurfaces ();

	gldepthmin = 0;
	gldepthmax = 0.5;
	glDepthRange (gldepthmin, gldepthmax);
	glDepthFunc (GL_LEQUAL);

	// blend on top
	glEnable (GL_BLEND);

	//mirror fix - from QER
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	//mirror fix

	glMatrixMode(GL_PROJECTION);
	if (mirror_plane->normal[2])
		glScalef (1,-1,1);
	else
		glScalef (-1,1,1);
	glCullFace(GL_FRONT);
	glMatrixMode(GL_MODELVIEW);

	glLoadMatrixf (r_base_world_matrix);

	glColor4f (1,1,1,r_mirroralpha.value);
	s = cl.worldmodel->textures[mirrortexturenum]->texturechain;
	for ( ; s ; s=s->texturechain)
		R_RenderBrushPoly (s);
	cl.worldmodel->textures[mirrortexturenum]->texturechain = NULL;
	glDisable (GL_BLEND);
	glColor4f (1,1,1,1);

	//mirror fix - from QER
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	//mirror fix
}

extern void R_Glare ();
extern void R_DrawGlare ();

/*
================
R_RenderView

r_refdef must be set before the first call
================
*/
void R_RenderView (void)
{
	double	time1, time2;
	GLfloat colors[4] = {(GLfloat) 0.0, (GLfloat) 0.0, (GLfloat) 1, (GLfloat) 0.20};

	if (r_norefresh.value)
		return;

	if (!r_worldentity.model || !cl.worldmodel)
		Sys_Error ("R_RenderView: NULL worldmodel");

	if (r_speeds.value)
	{
		glFinish ();
		time1 = Sys_FloatTime ();
		c_brush_polys = 0;
		c_alias_polys = 0;
	}

	mirror = false;

	if (gl_finish.value)
		glFinish ();

	R_Clear ();

	R_Glare ();

	// render normal view

/***** Experimental silly looking fog ******
****** Use r_fullbright if you enable ******
	glFogi(GL_FOG_MODE, GL_LINEAR);
	glFogfv(GL_FOG_COLOR, colors);
	glFogf(GL_FOG_END, 512.0);
	glEnable(GL_FOG);
********************************************/

	R_RenderScene ();
	R_DrawViewModel ();
	R_DrawWaterSurfaces ();

//  More fog right here :)
//	glDisable(GL_FOG);
//  End of all fog code...

	// render mirror view
	R_Mirror ();

	R_PolyBlend ();

	if (r_speeds.value)
	{
//		glFinish ();
		time2 = Sys_FloatTime ();
		Con_Printf ("%3i ms  %4i wpoly %4i epoly\n", (int)((time2-time1)*1000), c_brush_polys, c_alias_polys); 
	}
	R_DrawGlare ();
}


