/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// disable data conversion warnings

#pragma warning(disable : 4244)     // MIPS
#pragma warning(disable : 4136)     // X86
#pragma warning(disable : 4051)     // ALPHA
  
#ifdef _WIN32
#include <windows.h>
#endif

#include <GL/gl.h>
#include <GL/glu.h>

void GL_BeginRendering (int *x, int *y, int *width, int *height);
void GL_EndRendering (void);


#ifdef _WIN32
// Function prototypes for the Texture Object Extension routines
typedef GLboolean (APIENTRY *ARETEXRESFUNCPTR)(GLsizei, const GLuint *,
                    const GLboolean *);
typedef void (APIENTRY *BINDTEXFUNCPTR)(GLenum, GLuint);
typedef void (APIENTRY *DELTEXFUNCPTR)(GLsizei, const GLuint *);
typedef void (APIENTRY *GENTEXFUNCPTR)(GLsizei, GLuint *);
typedef GLboolean (APIENTRY *ISTEXFUNCPTR)(GLuint);
typedef void (APIENTRY *PRIORTEXFUNCPTR)(GLsizei, const GLuint *,
                    const GLclampf *);
typedef void (APIENTRY *TEXSUBIMAGEPTR)(int, int, int, int, int, int, int, int, void *);

extern	BINDTEXFUNCPTR bindTexFunc;
extern	DELTEXFUNCPTR delTexFunc;
extern	TEXSUBIMAGEPTR TexSubImage2DFunc;
#endif

extern	int texture_extension_number;
extern	int		texture_mode;

extern	float	gldepthmin, gldepthmax;

extern	int normcube_texture_object;//PENTA: gl texture object of the normalization cubemap 
extern  cvar_t	gl_watershader; //PENTA:
extern	float color_black[4];
extern	float	fog_color[4];

#define strncasecmp strnicmp

void GL_Upload32 (unsigned *data, int width, int height,  qboolean mipmap, qboolean alpha);
void GL_Upload8 (byte *data, int width, int height,  qboolean mipmap, qboolean alpha);
//==old code== int GL_LoadTexture (char *identifier, int width, int height, byte *data, qboolean mipmap, qboolean alpha, int bytesperpixel);
int GL_LoadTexture (char *identifier, int width, int height, byte *data, qboolean mipmap, qboolean alpha, int bytesperpixel);
//new code

int GL_FindTexture (char *identifier);
/*
typedef struct sh_glares
{
	int scr_fov; // attempted quick fix - Entar
	int glare; // attempted quick fix - Entar
	int gl_wireframe; // attempted quick fix - Entar
	int sh_glares; // attempted quick fix - Entar
};
*/

extern qboolean glare;

typedef struct surfcache_s
{
	struct surfcache_s	*next;
	struct surfcache_s 	**owner;		// NULL is an empty chunk of memory
	int					lightadj[MAXLIGHTMAPS]; // checked for strobe flush
	int					dlight;
	int					size;		// including header
	unsigned			width;
	unsigned			height;		// DEBUG only needed for debug
	float				mipscale;
	struct texture_s	*texture;	// checked for animating textures
	byte				data[4];	// width*height elements
} surfcache_t;

typedef union lightcmd_u {
  int asInt;
  float asFloat;
  //vec_t asVec;
  vec3_t asVec;
  void *asVoid;
} lightcmd_t;

typedef struct shadowlight_s
{
	vec3_t	origin;		//position of light source
	vec3_t	baseColor, color;		//light color, animated color
	float	radius;		//radius of light source (doesn't light anyting out of that circle)
						//so we can "clip" our shadowvolumes against that circle
	float	brightness;
	qboolean visible;	//light is "visible" this frame
	qboolean isStatic;	//light is static and has a precalc volume
	qboolean castShadow;//lights casts shadows
	qboolean halo;		//light has a halo
	mleaf_t	*leaf;		//leaf this light is in
	byte vis[MAX_MAP_LEAFS/8];//redone pvs for light, only poly's in nodes
	byte entvis[MAX_MAP_LEAFS/8];//original pvs a light origin
	msurface_t	**visSurf; //the surfaces that should cast shadows for this light (only when static)
	int		*volumeCmds;  //gl commands to draw the shadow volume
	float	*volumeVerts;
	int		numVolumeVerts;
	lightcmd_t	*lightCmds;	//gl commands to draw the cap/lighted volumes
	int		 numVisSurf;
	int		style;
	entity_t *owner;
	//screenrect_t scizz;
	int		filtercube;	//texture object of this light's cubemap filter
	vec3_t	angles; //angles of the cubemap filter
	float	rspeed; //rotation speed of cube map;
	float	cubescale; //scale factor of cube map;
	float	haloalpha; //alpha of halo
	vec3_t	oldlightorigin;
	
	//attempted fix - entar
	vec3_t	oldlightpos;

} shadowlight_t;

typedef struct brushlightinstant_s {

	int lockframe;	//number of the frame when it was last locked.
	vec3_t lasteorg, lastlorg, lasteangles, lastvorg;
	float lastlradius;
	entity_t	*lastent;
	shadowlight_t *lastlight;
	vec3_t	lightpos, vieworg; //Object space light position
	qboolean shadowonly, lastshadowonly;

	//Per light stuff
	vec3_t	extvertices[MAXALIASVERTS]; //extruded vertices (volumes)
	vec3_t	tslights[MAXALIASVERTS]; //per vertex tangent space light vector
	vec3_t	tshalfangles[MAXALIASVERTS]; //per vertex half angle vector
	float	atencoords[MAXALIASVERTS][2]; //per vertex attenuation texture coords for non gf3 cards
	float	colorscales[MAXALIASVERTS]; //per vertex attenuation texture coords for non gf3 cards
	qboolean polygonVis[MAXALIASTRIS]; //polygon facingness for lights
	qboolean neighbourVis[MAXALIASTRIS]; //polygon facingness for lights
} brushlightinstant_t;

//PENTA: new
typedef struct aliaslightinstant_s {

	int lockframe;	//number of the frame when it was last locked.
	vec3_t lasteorg, lastlorg, lasteangles, lastvorg;
	float lastlradius;
	entity_t	*lastent;
	aliashdr_t * lasthdr;
	shadowlight_t *lastlight;
	vec3_t	lightpos, vieworg; //Object space light position
	void *lastframeinstant;

	//Per light stuff
	vec3_t	extvertices[MAXALIASVERTS]; //extruded vertices (volumes)
	vec3_t	tslights[MAXALIASVERTS]; //per vertex tangent space light vector
	vec3_t	tshalfangles[MAXALIASVERTS]; //per vertex half angle vector
	vec3_t	colors[MAXALIASVERTS]; //per vertex attenuation for non gf3 cards
	qboolean triangleVis[MAXALIASTRIS]; //triangle facingness for lights
	
	int		indecies[MAXALIASTRIS*3];//light visibility list
	int numtris;
	
} aliaslightinstant_t;

typedef struct aliasframeinstant_s {
	
     // chain of frameinstant for multiple surfaces
        struct aliasframeinstant_s *_next;
	
	int lockframe;	//number of the frame when it was last locked.
	int	updateframe; //number of the fame when it was last recalculated
	aliashdr_t *paliashdr;
	int lastpose1, lastpose2;
	float lastblend;
	entity_t	*lastent;
	qboolean lastshadowonly;
	aliashdr_t *lastpaliashdr;

	qboolean	shadowonly; //We only need the instant for shadow volume drawing
							//we use this for ents out of the frustum but e.g. behind the player
	//Per entity stuff
	vec3_t	vertices[MAXALIASVERTS]; //interpolated vertices
	vec3_t	normals[MAXALIASVERTS]; //interpolated vertex normals
	vec3_t	tangents[MAXALIASVERTS]; //interpolated vertex tangents
	vec3_t	binomials[MAXALIASVERTS]; //interpolated vertex binomials
	plane_t	triplanes[MAXALIASTRIS]; //triangle plane equatations

	aliaslightinstant_t *lightinstant;
} aliasframeinstant_t;

typedef struct
{
	float	x, y, z;
	float	s, t;
	float	r, g, b;
} glvert_t;

static int particlepalette[256] =
{
	0x000000,0x0f0f0f,0x1f1f1f,0x2f2f2f,0x3f3f3f,0x4b4b4b,0x5b5b5b,0x6b6b6b,
	0x7b7b7b,0x8b8b8b,0x9b9b9b,0xababab,0xbbbbbb,0xcbcbcb,0xdbdbdb,0xebebeb,
	0x0f0b07,0x170f0b,0x1f170b,0x271b0f,0x2f2313,0x372b17,0x3f2f17,0x4b371b,
	0x533b1b,0x5b431f,0x634b1f,0x6b531f,0x73571f,0x7b5f23,0x836723,0x8f6f23,
	0x0b0b0f,0x13131b,0x1b1b27,0x272733,0x2f2f3f,0x37374b,0x3f3f57,0x474767,
	0x4f4f73,0x5b5b7f,0x63638b,0x6b6b97,0x7373a3,0x7b7baf,0x8383bb,0x8b8bcb,
	0x000000,0x070700,0x0b0b00,0x131300,0x1b1b00,0x232300,0x2b2b07,0x2f2f07,
	0x373707,0x3f3f07,0x474707,0x4b4b0b,0x53530b,0x5b5b0b,0x63630b,0x6b6b0f,
	0x070000,0x0f0000,0x170000,0x1f0000,0x270000,0x2f0000,0x370000,0x3f0000,
	0x470000,0x4f0000,0x570000,0x5f0000,0x670000,0x6f0000,0x770000,0x7f0000,
	0x131300,0x1b1b00,0x232300,0x2f2b00,0x372f00,0x433700,0x4b3b07,0x574307,
	0x5f4707,0x6b4b0b,0x77530f,0x835713,0x8b5b13,0x975f1b,0xa3631f,0xaf6723,
	0x231307,0x2f170b,0x3b1f0f,0x4b2313,0x572b17,0x632f1f,0x733723,0x7f3b2b,
	0x8f4333,0x9f4f33,0xaf632f,0xbf772f,0xcf8f2b,0xdfab27,0xefcb1f,0xfff31b,
	0x0b0700,0x1b1300,0x2b230f,0x372b13,0x47331b,0x533723,0x633f2b,0x6f4733,
	0x7f533f,0x8b5f47,0x9b6b53,0xa77b5f,0xb7876b,0xc3937b,0xd3a38b,0xe3b397,
	0xab8ba3,0x9f7f97,0x937387,0x8b677b,0x7f5b6f,0x775363,0x6b4b57,0x5f3f4b,
	0x573743,0x4b2f37,0x43272f,0x371f23,0x2b171b,0x231313,0x170b0b,0x0f0707,
	0xbb739f,0xaf6b8f,0xa35f83,0x975777,0x8b4f6b,0x7f4b5f,0x734353,0x6b3b4b,
	0x5f333f,0x532b37,0x47232b,0x3b1f23,0x2f171b,0x231313,0x170b0b,0x0f0707,
	0xdbc3bb,0xcbb3a7,0xbfa39b,0xaf978b,0xa3877b,0x977b6f,0x876f5f,0x7b6353,
	0x6b5747,0x5f4b3b,0x533f33,0x433327,0x372b1f,0x271f17,0x1b130f,0x0f0b07,
	0x6f837b,0x677b6f,0x5f7367,0x576b5f,0x4f6357,0x475b4f,0x3f5347,0x374b3f,
	0x2f4337,0x2b3b2f,0x233327,0x1f2b1f,0x172317,0x0f1b13,0x0b130b,0x070b07,
	0xfff31b,0xefdf17,0xdbcb13,0xcbb70f,0xbba70f,0xab970b,0x9b8307,0x8b7307,
	0x7b6307,0x6b5300,0x5b4700,0x4b3700,0x3b2b00,0x2b1f00,0x1b0f00,0x0b0700,
	0x0000ff,0x0b0bef,0x1313df,0x1b1bcf,0x2323bf,0x2b2baf,0x2f2f9f,0x2f2f8f,
	0x2f2f7f,0x2f2f6f,0x2f2f5f,0x2b2b4f,0x23233f,0x1b1b2f,0x13131f,0x0b0b0f,
	0x2b0000,0x3b0000,0x4b0700,0x5f0700,0x6f0f00,0x7f1707,0x931f07,0xa3270b,
	0xb7330f,0xc34b1b,0xcf632b,0xdb7f3b,0xe3974f,0xe7ab5f,0xefbf77,0xf7d38b,
	0xa77b3b,0xb79b37,0xc7c337,0xe7e357,0x7fbfff,0xabe7ff,0xd7ffff,0x670000,
	0x8b0000,0xb30000,0xd70000,0xff0000,0xfff393,0xfff7c7,0xffffff,0x9f5b53
};

static const int tex_particle = 63;

extern glvert_t glv;

extern	int glx, gly, glwidth, glheight;

#ifdef _WIN32
extern	PROC glArrayElementEXT;
extern	PROC glColorPointerEXT;
extern	PROC glTexturePointerEXT;
extern	PROC glVertexPointerEXT;
#endif

#define ALIAS_BASE_SIZE_RATIO		(1.0 / 11.0)
					// normalizing factor so player model works out to about
					//  1 pixel per triangle
#define	MAX_LBM_HEIGHT		480

#define TILE_SIZE		128		// size of textures generated by R_GenTiledSurf

#define SKYSHIFT		7
#define	SKYSIZE			(1 << SKYSHIFT)
#define SKYMASK			(SKYSIZE - 1)

#define BACKFACE_EPSILON	0.01


void R_TimeRefresh_f (void);
void R_ReadPointFile_f (void);
texture_t *R_TextureAnimation (texture_t *base);

typedef enum {
	pt_static, pt_grav, pt_slowgrav, pt_fire, pt_explode, pt_explode2, pt_blob, pt_blob2
} ptype_t;

//ParticleEffect_t *effect;
//void		R_DrawDecals (void);
//void		R_InitDecals (void);
//extern void R_SpawnDecal(vec3_t center, vec3_t normal, vec3_t tangent,  ParticleEffect_t *effect);
//void		R_SpawnDecal(vec3_t center, vec3_t normal, vec3_t tangent);
//void		R_ClearDecals(void);

float RandomMinMax(float min, float max);

typedef enum {
	emt_box
} emtype_t;

typedef enum {
	align_view, align_vel, align_surf
} align_t;

typedef struct ParticleEffect_s {
	char name[64];
	vec3_t emmiterParams1, emmiterParams2; //mins maxs of emmiter box
	emtype_t emmiterType; //currently only box is supported
	vec3_t startcolormin, startcolormax, endcolormin, endcolormax;
	vec3_t velocitymin, velocitymax; //min max velocity
	float	lifemin, lifemax; //min max life
	int	movetype; //how it moves (gravity/collision...)
	vec3_t	gravity;
	float	rotmin, rotmax;//rotaiton speed
	float	growmin, growmax;//scale speed (make second smaller to shrink)
	float	sizemin, sizemax;//scale speed (make second smaller to shrink)
	vec3_t	drag; //drag of particle (how fast it looses it's speed)
	int	srcblend, dstblend; // gl enums for blend modes
	int	numbounces; //number of bounces before particle is deleted
	int	texture; //gl object of the texture	
	align_t	align;//particle is aligned with its velocity
	float	velscale;
	struct ParticleEffect_s *spawn; //particle effect to spawn on hit
	struct ParticleEffect_s *next;
	struct ParticleEffect_s *effect;
} ParticleEffect_t;

typedef enum {
	dt_blood
} dtype_t;

extern qboolean gl_var;
#define MAX_VOLUME_COMMANDS 131072 //Thats 0.5 meg storage for commands, insane
#define MAX_VOLUME_VERTS 87381 //1 Meg storage for vertices
#define AGP_BUFFER_SIZE 2048*4
extern void *AGP_Buffer;
extern  cvar_t  sh_worldshadows;

#define MAX_DECAL_VERTICES 128
#define MAX_DECAL_TRIANGLES 64
//PENTA: Decals
typedef struct decal_s
{
	vec3_t		origin;
	vec3_t		normal;
	vec3_t		tangent;
	float		radius;

	float		color[4], startcolor[4], endcolor[4];
	struct decal_s	*next;
	float		die;
	float		lifetime;
	dtype_t		type;
	int			texture; //texture object of particle
	int			srcblend;
	int			dstblend;

	//geometry of decail
	int	vertexCount, triangleCount;
	vec3_t		vertexArray[MAX_DECAL_VERTICES];
	float		texcoordArray[MAX_DECAL_VERTICES][2];
	int			triangleArray[MAX_DECAL_TRIANGLES][3];
} decal_t;

typedef enum
{
	GENERIC = 0,
	GEFORCE,
	GEFORCE3,
	RADEON,
	PARHELIA,
	ARB
} qcardtype;
extern qcardtype gl_cardtype;

// !!! if this is changed, it must be changed in d_ifacea.h too !!!
/*typedef struct particle_s
{
// driver-usable fields
	vec3_t		org;
	float		color;
// drivers never touch the following fields
	struct particle_s	*next;
	vec3_t		vel;
	float		ramp;
	float		die;
	ptype_t		type;
} particle_t;
*/
//====================================================

extern	entity_t	r_worldentity;
extern	qboolean	r_cache_thrash;		// compatability
extern	vec3_t		modelorg, r_entorigin;
extern	entity_t	*currententity;
extern	int			r_visframecount;	// ??? what difs?
extern	int			r_framecount;
extern	mplane_t	frustum[4];
extern	int		c_brush_polys, c_alias_polys;


//
// view origin
//
extern	vec3_t	vup;
extern	vec3_t	vpn;
extern	vec3_t	vright;
extern	vec3_t	r_origin;

//
// screen size info
//
extern	refdef_t	r_refdef;
extern	mleaf_t		*r_viewleaf, *r_oldviewleaf;
extern	texture_t	*r_notexture_mip;
extern	int		d_lightstylevalue[256];	// 8.8 fraction of base light value

extern	qboolean	envmap;
extern	int	currenttexture;
extern	int	cnttextures[2];
extern	int	particletexture;
extern	int	playertextures;

extern	int	skytexturenum;		// index in cl.loadmodel, not gl texture object

//extern	int	underwatertexture;	// shinetex_glass, shinetex_chrome, underwatertexture;
extern	int shinetex_glass, shinetex_chrome, underwatertexture;

extern	cvar_t	sh_glares; // glares stuff
extern  cvar_t	gl_wireframe; //
extern  cvar_t	scr_fov; //
extern	cvar_t	cl_sbar;
extern	cvar_t	gl_dark;
extern	cvar_t	gl_particles; // new particles cvar

#define MAX_LIGHT_COMMANDS 65536 //0.25 meg for light commands

extern  cvar_t	sh_noshadowpopping;
extern	cvar_t	sh_colormaps;//PENTA: enable disable textures on the world (light will remain)
extern	int	glow_texture_object;	//PENTA: gl texture object of the glow texture
extern	shadowlight_t *currentshadowlight;
extern  mmvertex_t *globalVertexTable;
extern  lightcmd_t	lightCmdsBuff[MAX_LIGHT_COMMANDS+128];
extern	data;

int		indecies[MAXALIASTRIS*3];//light visibility list

void	R_WorldToObjectMatrix(entity_t *e, matrix_4x4 result);
void		R_DrawAliasFrameWV (aliashdr_t *paliashdr, aliasframeinstant_t *instant, qboolean specular);
void		R_DrawAliasObjectLight (entity_t *e, void (*AliasGeoSender) (aliashdr_t *paliashdr, aliasframeinstant_t* instant));

extern	cvar_t	r_norefresh;
extern	cvar_t	r_drawentities;
extern	cvar_t	r_drawworld;
extern	cvar_t	r_drawviewmodel;
extern	cvar_t	r_speeds;
extern	cvar_t	r_waterwarp;
extern	cvar_t	r_fullbright;
extern	cvar_t	r_lightmap;
extern	cvar_t	r_shadows;

extern cvar_t gl_asynch;
extern cvar_t r_interpolations; //interpolation stuff

extern	cvar_t	r_mirroralpha;
extern	cvar_t	r_wateralpha;
extern	cvar_t	r_dynamic;
extern	cvar_t	r_novis;

// fenix@io.com: model interpolation
//extern  cvar_t  r_interpolate_model_animation;
extern  cvar_t  r_interpolate_model_a;
//extern  cvar_t  r_interpolate_model_transform;
extern  cvar_t  r_interpolate_model_t;

extern	cvar_t	gl_clear;
extern	cvar_t	gl_cull;
extern	cvar_t	gl_poly;
extern	cvar_t	gl_texsort;
extern	cvar_t	gl_smoothmodels;
extern	cvar_t	gl_affinemodels;
extern	cvar_t	gl_polyblend;
extern	cvar_t	gl_keeptjunctions;
extern	cvar_t	gl_reporttjunctions;
extern	cvar_t	gl_flashblend;
extern	cvar_t	gl_nocolors;
extern	cvar_t	gl_doubleeyes;

extern	int		gl_lightmap_format;
extern	int		gl_solid_format;
extern	int		gl_alpha_format;

extern	cvar_t	gl_max_size;
extern	cvar_t	gl_playermip;

extern	cvar_t	gl_caustics;
extern	cvar_t	gl_shiny;
extern  cvar_t  gl_detail;
extern  cvar_t  gl_dualwater;
extern	cvar_t	gl_rainparticles;
extern	cvar_t	gl_rainparticles_quality;

extern	int			mirrortexturenum;	// quake texturenum, not gltexturenum
extern	qboolean	mirror;
extern	mplane_t	*mirror_plane;

extern	float	r_world_matrix[16];

extern	const char *gl_vendor;
extern	const char *gl_renderer;
extern	const char *gl_version;
extern	const char *gl_extensions;

void R_TranslatePlayerSkin (int playernum);
void GL_Bind (int texnum);

// Multitexture
#define    TEXTURE0_SGIS				0x835E
#define    TEXTURE1_SGIS				0x835F

#ifndef _WIN32
#define APIENTRY /* */
#endif

typedef void (APIENTRY *lpMTexFUNC) (GLenum, GLfloat, GLfloat);
typedef void (APIENTRY *lpSelTexFUNC) (GLenum);
extern lpMTexFUNC qglMTexCoord2fSGIS;
extern lpSelTexFUNC qglSelectTextureSGIS;

extern qboolean gl_mtexable;

void GL_DisableMultitexture(void);
void GL_EnableMultitexture(void);

//
#define Q3MODELS 1      //Allow Q3 models in the same way as Q1 models. --> JTR 
void   Mod_LoadMd3Model (model_t *mod, void *buffer); 
//

void   R_DrawGlare (void);
void   R_Glare (void);
void   R_InitGlare (void);

/* ARB_multitexture defines and prototypes from <GL/gl.h> */
#define GL_ACTIVE_TEXTURE_ARB               0x84E0
#define GL_CLIENT_ACTIVE_TEXTURE_ARB        0x84E1
#define GL_MAX_ACTIVE_TEXTURES_ARB          0x84E2
#define GL_TEXTURE0_ARB                     0x84C0
#define GL_TEXTURE1_ARB                     0x84C1
#define GL_TEXTURE2_ARB                     0x84C2

//PENTA: Texture env combine/ Texture env dot3
#define GL_COMBINE_ARB                    0x8570
#define GL_COMBINE_RGB_ARB                0x8571
#define GL_COMBINE_ALPHA_ARB              0x8572
#define GL_RGB_SCALE_ARB                  0x8573
#define GL_ADD_SIGNED_ARB                 0x8574
#define GL_INTERPOLATE_ARB                0x8575
#define GL_SUBTRACT_ARB                   0x84E7
#define GL_CONSTANT_ARB                   0x8576
#define GL_PRIMARY_COLOR_ARB              0x8577
#define GL_PREVIOUS_ARB                   0x8578
#define GL_SOURCE0_RGB_ARB                0x8580
#define GL_SOURCE1_RGB_ARB                0x8581
#define GL_SOURCE2_RGB_ARB                0x8582
#define GL_SOURCE3_RGB_ARB                0x8583
#define GL_SOURCE4_RGB_ARB                0x8584
#define GL_SOURCE5_RGB_ARB                0x8585
#define GL_SOURCE6_RGB_ARB                0x8586
#define GL_SOURCE7_RGB_ARB                0x8587
#define GL_SOURCE0_ALPHA_ARB              0x8588
#define GL_SOURCE1_ALPHA_ARB              0x8589
#define GL_SOURCE2_ALPHA_ARB              0x858A
#define GL_SOURCE3_ALPHA_ARB              0x858B
#define GL_SOURCE4_ALPHA_ARB              0x858C
#define GL_SOURCE5_ALPHA_ARB              0x858D
#define GL_SOURCE6_ALPHA_ARB              0x858E
#define GL_SOURCE7_ALPHA_ARB              0x858F
#define GL_OPERAND0_RGB_ARB               0x8590
#define GL_OPERAND1_RGB_ARB               0x8591
#define GL_OPERAND2_RGB_ARB               0x8592
#define GL_OPERAND3_RGB_ARB               0x8593
#define GL_OPERAND4_RGB_ARB               0x8594
#define GL_OPERAND5_RGB_ARB               0x8595
#define GL_OPERAND6_RGB_ARB               0x8596
#define GL_OPERAND7_RGB_ARB               0x8597
#define GL_OPERAND0_ALPHA_ARB             0x8598
#define GL_OPERAND1_ALPHA_ARB             0x8599
#define GL_OPERAND2_ALPHA_ARB             0x859A
#define GL_OPERAND3_ALPHA_ARB             0x859B
#define GL_OPERAND4_ALPHA_ARB             0x859C
#define GL_OPERAND5_ALPHA_ARB             0x859D
#define GL_OPERAND6_ALPHA_ARB             0x859E
#define GL_OPERAND7_ALPHA_ARB             0x859F

/*
#define	GL_DOT3_RGB_EXT					  0x8740
#define	GL_DOT3_RGBA_EXT				  0x8741
*/
#define GL_DOT3_RGB_ARB					0x86AE
#define GL_DOT3_RGBA_ARB				0x86AF


//PENTA: Cube mapping
#define GL_NORMAL_MAP_ARB                 0x8511
#define GL_REFLECTION_MAP_ARB             0x8512
#define GL_TEXTURE_CUBE_MAP_ARB           0x8513
#define GL_TEXTURE_BINDING_CUBE_MAP_ARB   0x8514
#define GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB 0x8515
#define GL_TEXTURE_CUBE_MAP_NEGATIVE_X_ARB 0x8516
#define GL_TEXTURE_CUBE_MAP_POSITIVE_Y_ARB 0x8517
#define GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_ARB 0x8518
#define GL_TEXTURE_CUBE_MAP_POSITIVE_Z_ARB 0x8519
#define GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_ARB 0x851A
#define GL_PROXY_TEXTURE_CUBE_MAP_ARB     0x851B
#define GL_MAX_CUBE_MAP_TEXTURE_SIZE_ARB  0x851C

//PENTA: NV register combiners
#define GL_REGISTER_COMBINERS_NV          0x8522
#define GL_VARIABLE_A_NV                  0x8523
#define GL_VARIABLE_B_NV                  0x8524
#define GL_VARIABLE_C_NV                  0x8525
#define GL_VARIABLE_D_NV                  0x8526
#define GL_VARIABLE_E_NV                  0x8527
#define GL_VARIABLE_F_NV                  0x8528
#define GL_VARIABLE_G_NV                  0x8529
#define GL_CONSTANT_COLOR0_NV             0x852A
#define GL_CONSTANT_COLOR1_NV             0x852B
#define GL_PRIMARY_COLOR_NV               0x852C
#define GL_SECONDARY_COLOR_NV             0x852D
#define GL_SPARE0_NV                      0x852E
#define GL_SPARE1_NV                      0x852F
#define GL_DISCARD_NV                     0x8530
#define GL_E_TIMES_F_NV                   0x8531
#define GL_SPARE0_PLUS_SECONDARY_COLOR_NV 0x8532
#define GL_UNSIGNED_IDENTITY_NV           0x8536
#define GL_UNSIGNED_INVERT_NV             0x8537
#define GL_EXPAND_NORMAL_NV               0x8538
#define GL_EXPAND_NEGATE_NV               0x8539
#define GL_HALF_BIAS_NORMAL_NV            0x853A
#define GL_HALF_BIAS_NEGATE_NV            0x853B
#define GL_SIGNED_IDENTITY_NV             0x853C
#define GL_SIGNED_NEGATE_NV               0x853D
#define GL_SCALE_BY_TWO_NV                0x853E
#define GL_SCALE_BY_FOUR_NV               0x853F
#define GL_SCALE_BY_ONE_HALF_NV           0x8540
#define GL_BIAS_BY_NEGATIVE_ONE_HALF_NV   0x8541
#define GL_COMBINER_INPUT_NV              0x8542
#define GL_COMBINER_MAPPING_NV            0x8543
#define GL_COMBINER_COMPONENT_USAGE_NV    0x8544
#define GL_COMBINER_AB_DOT_PRODUCT_NV     0x8545
#define GL_COMBINER_CD_DOT_PRODUCT_NV     0x8546
#define GL_COMBINER_MUX_SUM_NV            0x8547
#define GL_COMBINER_SCALE_NV              0x8548
#define GL_COMBINER_BIAS_NV               0x8549
#define GL_COMBINER_AB_OUTPUT_NV          0x854A
#define GL_COMBINER_CD_OUTPUT_NV          0x854B
#define GL_COMBINER_SUM_OUTPUT_NV         0x854C
#define GL_MAX_GENERAL_COMBINERS_NV       0x854D
#define GL_NUM_GENERAL_COMBINERS_NV       0x854E
#define GL_COLOR_SUM_CLAMP_NV             0x854F
#define GL_COMBINER0_NV                   0x8550
#define GL_COMBINER1_NV                   0x8551
#define GL_COMBINER2_NV                   0x8552
#define GL_COMBINER3_NV                   0x8553
#define GL_COMBINER4_NV                   0x8554
#define GL_COMBINER5_NV                   0x8555
#define GL_COMBINER6_NV                   0x8556
#define GL_COMBINER7_NV                   0x8557

#define GL_TEXTURE_MAX_ANISOTROPY_EXT		0x84FE
#define GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT	0x84FF

