/* 
Copyright (C) 1997-2001 Id Software, Inc. 

This program is free software; you can redistribute it and/or 
modify it under the terms of the GNU General Public License 
as published by the Free Software Foundation; either version 2 
of the License, or (at your option) any later version. 

This program is distributed in the hope that it will be useful, 
but WITHOUT ANY WARRANTY; without even the implied warranty of 
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details. 

You should have received a copy of the GNU General Public License 
along with this program; if not, write to the Free Software 
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. 

*/ 

// r_postprocess.cpp -> post process effects (motion blur,glows) 

#include "r_local.h" 

void DrawQuad(int width, int height, int ofsX, int ofsY) 
{ 
    qglTexCoord2f(ofsX,ofsY); 
    qglVertex3f(0, 1, -1); 
    qglTexCoord2f(ofsX,height+ofsY); 
    qglVertex3f(0, 0, -1); 
    qglTexCoord2f(width+ofsX,height+ofsY); 
    qglVertex3f(1, 0, -1); 
    qglTexCoord2f(width+ofsX,ofsY); 
    qglVertex3f(1, 1, -1); 
} 

void R_RenderPostProcess(int type) 
{ 
   if (type == 1) // PP_GLOW 
   { 
      // a simple check to see if the blur is enabled 
      if (r_glow_blur_steps->value == 0 || r_glow_strength->value == 0) 
         return; 

      // enable some OpenGL stuff 
      qglEnable(GL_TEXTURE_RECTANGLE_NV);      
      qglColor3f(1,1,1); 
      qglDisable(GL_DEPTH_TEST); 

      // Grab the screen and put it into the Screen texture 
      GL_Bind(0, r_screentexture); 
      qglCopyTexImage2D(GL_TEXTURE_RECTANGLE_NV, 0, GL_RGB, 0, 0, r_refdef.width, r_refdef.height, 0); 

      // Set up an orthogonal projection 
      qglMatrixMode(GL_MODELVIEW); 
      qglPushMatrix(); 
      qglLoadIdentity(); 
      
      qglMatrixMode(GL_PROJECTION); 
      qglPushMatrix(); 
      qglLoadIdentity(); 
      qglOrtho(0, 1, 1, 0, 0.1, 100); // a 1 unit size of view for DrawQuad to draw to 

      // Render the current scene to a new, lower-res texture, darkening non-bright areas of the scene 
      // by multiplying it with itself a few times. 

      qglViewport(0, 0, r_refdef.width * 0.5, r_refdef.height * 0.5); 

      GL_Bind(0, r_screentexture); 

      qglBlendFunc(GL_DST_COLOR, GL_ZERO); 
      
      qglDisable(GL_BLEND); 

      qglBegin(GL_QUADS); 
      DrawQuad(r_refdef.width, r_refdef.height); 
      qglEnd(); 

      qglEnable(GL_BLEND); 

      qglBegin(GL_QUADS); 
      for (int i = 0; i < r_glow_darken_steps->value; i++) 
         DrawQuad(r_refdef.width, r_refdef.height); 
      qglEnd(); 

      GL_Bind(0, r_effecttexture); 

      qglCopyTexImage2D(GL_TEXTURE_RECTANGLE_NV, 0, GL_RGB, 0, 0, r_refdef.width * 0.5, r_refdef.height * 0.5, 0); 

      // Blur the now darkened scene in the horizontal direction. 

      float blurAlpha = 1 / ((r_glow_blur_steps->value * 2) + 1); 

      qglColor4f(1,1,1,blurAlpha); 

      qglBlendFunc(GL_SRC_ALPHA, GL_ZERO); 

      qglBegin(GL_QUADS); 
      DrawQuad(r_refdef.width * 0.5, r_refdef.height * 0.5); 
      qglEnd(); 

      qglBlendFunc(GL_SRC_ALPHA, GL_ONE); 

      qglBegin(GL_QUADS); 
      for (i = 1; i <= r_glow_blur_steps->value; i++) 
      { 
         DrawQuad(r_refdef.width * 0.5, r_refdef.height * 0.5, -i, 0); 
         DrawQuad(r_refdef.width * 0.5, r_refdef.height * 0.5, i, 0); 
      } 
      qglEnd(); 
      
      qglCopyTexImage2D(GL_TEXTURE_RECTANGLE_NV, 0, GL_RGB, 0, 0, r_refdef.width * 0.5, r_refdef.height * 0.5, 0); 

      // Blur the horizontally blurred image in the vertical direction. 

      qglBlendFunc(GL_SRC_ALPHA, GL_ZERO); 

      qglBegin(GL_QUADS); 
      DrawQuad(r_refdef.width * 0.5, r_refdef.height * 0.5); 
      qglEnd(); 

      qglBlendFunc(GL_SRC_ALPHA, GL_ONE); 

      qglBegin(GL_QUADS); 
      for (i = 1; i <= r_glow_blur_steps->value; i++) 
      { 
         DrawQuad(r_refdef.width * 0.5, r_refdef.height * 0.5, 0, -i); 
         DrawQuad(r_refdef.width * 0.5, r_refdef.height * 0.5, 0, i); 
      } 
      qglEnd(); 
      
      qglCopyTexImage2D(GL_TEXTURE_RECTANGLE_NV, 0, GL_RGB, 0, 0, r_refdef.width * 0.5, r_refdef.height * 0.5, 0); 

      // Combine the blur with the original image. 

      qglViewport(0, 0, r_refdef.width, r_refdef.height); 

       qglDisable(GL_BLEND); 

      qglBegin(GL_QUADS); 
      DrawQuad(r_refdef.width * 0.5, r_refdef.height * 0.5); 
      qglEnd(); 

      qglEnable(GL_BLEND); 
      qglBlendFunc(GL_ONE, GL_ONE); 

      qglBegin(GL_QUADS); 
       for (i = 1; i < r_glow_strength->value; i++) 
      { 
         DrawQuad(r_refdef.width * 0.5, r_refdef.height * 0.5); 
      } 
      qglEnd(); 

      GL_Bind(0, r_screentexture); 

      qglBegin(GL_QUADS); 
      DrawQuad(r_refdef.width, r_refdef.height); 
      qglEnd(); 

      // Restore the original projection and modelview matrices and disable rectangular textures. 

      qglMatrixMode(GL_PROJECTION); 
      qglPopMatrix(); 

      qglMatrixMode(GL_MODELVIEW); 
      qglPopMatrix(); 

       qglDisable(GL_TEXTURE_RECTANGLE_NV); 
       qglEnable(GL_DEPTH_TEST); 
      qglDisable(GL_BLEND); 
   } 
}