/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// sbar.c -- status bar code

#include "quakedef.h"


int			sb_updates;		// if >= vid.numpages, no update needed

#define STAT_MINUS		10	// num frame for '-' stats digit
qpic_t		*sb_nums[2][11];
qpic_t		*sb_colon, *sb_slash;
qpic_t		*sb_ibar;
qpic_t		*sb_sbar;
qpic_t		*sb_scorebar;

qpic_t      *sb_weapons[7][8];   // 0 is active, 1 is owned, 2-5 are flashes
qpic_t      *sb_ammo[4];
qpic_t		*sb_sigil[4];
qpic_t		*sb_armor[3];
qpic_t		*sb_items[32];

qpic_t	*sb_faces[7][2];		// 0 is gibbed, 1 is dead, 2-6 are alive
							// 0 is static, 1 is temporary animation
qpic_t	*sb_face_invis;
qpic_t	*sb_face_quad;
qpic_t	*sb_face_invuln;
qpic_t	*sb_face_invis_invuln;

qboolean	sb_showscores;

int			sb_lines;			// scan lines to draw

qpic_t      *rsb_invbar[2];
qpic_t      *rsb_weapons[5];
qpic_t      *rsb_items[2];
qpic_t      *rsb_ammo[3];
qpic_t      *rsb_teambord;		// PGM 01/19/97 - team color border

//MED 01/04/97 added two more weapons + 3 alternates for grenade launcher
qpic_t      *hsb_weapons[7][5];   // 0 is active, 1 is owned, 2-5 are flashes
//MED 01/04/97 added array to simplify weapon parsing
int         hipweapons[4] = {HIT_LASER_CANNON_BIT,HIT_MJOLNIR_BIT,4,HIT_PROXIMITY_GUN_BIT};
//MED 01/04/97 added hipnotic items array
qpic_t      *hsb_items[2];

void Sbar_MiniDeathmatchOverlay (void);
void Sbar_DeathmatchOverlay (void);
void M_DrawPic (int x, int y, qpic_t *pic);


extern int addLeafs;
extern int addSurfs;

cvar_t r_hud = {"r_hud", "1", true};

/*
===============
Sbar_ShowScores

Tab key down
===============
*/
void Sbar_ShowScores (void)
{
	// toggle on/off rather than only on when down
	sb_showscores = !sb_showscores;
//	if (sb_showscores)
//		return;
//	sb_showscores = true;
	sb_updates = 0;
}

/*
===============
Sbar_DontShowScores

Tab key up
===============
*/
void Sbar_DontShowScores (void)
{
//	sb_showscores = false;
//	sb_updates = 0;
}

/*
===============
Sbar_Changed
===============
*/
void Sbar_Changed (void)
{
	sb_updates = 0;	// update next frame
}



/*
===============
GL Quake crosshair
===============
*/
int xhairtexture;

int			xhair_xpos;
int			xhair_ypos;


// smaller crosshair picture for resolutions of 640 x 480 and 800 x 600
unsigned char xhairpic[16][16] =
{
	{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
	{0,0,0,0,0,1,1,0,0,0,1,1,0,0,0,0},
	{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
};


// larger crosshair picture for resolutions of 1024 x 768 and higher
unsigned char xhairpic_hires[16][16] =
{
	{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
	{0,0,0,1,1,1,0,0,0,0,1,1,1,0,0,0},
	{0,0,0,1,1,1,0,0,0,0,1,1,1,0,0,0},
	{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
};


void R_InitXHairTexture (void)
{
#ifdef GLQUAKE
	int x, y;
	byte	data[16][16][4];
	float	xhair_xcoord_mody, xhair_ycoord_mody;

	xhairtexture = texture_extension_number++;
	GL_Bind (xhairtexture);

	for (x=0 ; x<16 ; x++)
	{
		for (y=0 ; y<16 ; y++)
		{
			// make it green
			data[y][x][0] = 0;
			data[y][x][1] = 255;
			data[y][x][2] = 0;
			if (vid.width <= 800)
				data[y][x][3] = xhairpic[x][y] * 255;
			else data[y][x][3] = xhairpic_hires[x][y] * 255;
		}
	}

	glTexImage2D (GL_TEXTURE_2D, 0, gl_alpha_format, 16, 16, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	xhair_xpos = (vid.width / 2) - 8;
	xhair_ypos = (vid.height / 2) + 8;
#endif
}


// draw the crosshair from the sbar rather than as normal... this way, we can ignore it 
// properly if we have the axe
void Draw_XHair (int x, int y)
{
#ifdef GLQUAKE
	glEnable (GL_BLEND);
	GL_Bind (xhairtexture);

	glBegin (GL_QUADS);

	glTexCoord2f (0, 0);
	glVertex2f (x, y);

	glTexCoord2f (1, 0);
	glVertex2f (x+16, y);

	glTexCoord2f (1, 1);
	glVertex2f (x+16, y+16);

	glTexCoord2f (0, 1);
	glVertex2f (x, y+16);

	glEnd ();
	glDisable (GL_BLEND);
#else
#endif
}


/*
===============
Sbar_Init
===============
*/
void Sbar_Init (void)
{
	int		i;

	for (i=0 ; i<10 ; i++)
	{
		sb_nums[0][i] = Draw_PicFromWad (va("num_%i",i));
		sb_nums[1][i] = Draw_PicFromWad (va("anum_%i",i));
	}

	Cvar_RegisterVariable (&r_hud);

	sb_nums[0][10] = Draw_PicFromWad ("num_minus");
	sb_nums[1][10] = Draw_PicFromWad ("anum_minus");

	sb_colon = Draw_PicFromWad ("num_colon");
	sb_slash = Draw_PicFromWad ("num_slash");

	sb_weapons[0][0] = Draw_PicFromWad ("inv_shotgun");
	sb_weapons[0][1] = Draw_PicFromWad ("inv_sshotgun");
	sb_weapons[0][2] = Draw_PicFromWad ("inv_nailgun");
	sb_weapons[0][3] = Draw_PicFromWad ("inv_snailgun");
	sb_weapons[0][4] = Draw_PicFromWad ("inv_rlaunch");
	sb_weapons[0][5] = Draw_PicFromWad ("inv_srlaunch");
	sb_weapons[0][6] = Draw_PicFromWad ("inv_lightng");

	sb_weapons[1][0] = Draw_PicFromWad ("inv2_shotgun");
	sb_weapons[1][1] = Draw_PicFromWad ("inv2_sshotgun");
	sb_weapons[1][2] = Draw_PicFromWad ("inv2_nailgun");
	sb_weapons[1][3] = Draw_PicFromWad ("inv2_snailgun");
	sb_weapons[1][4] = Draw_PicFromWad ("inv2_rlaunch");
	sb_weapons[1][5] = Draw_PicFromWad ("inv2_srlaunch");
	sb_weapons[1][6] = Draw_PicFromWad ("inv2_lightng");

	for (i=0 ; i<5 ; i++)
	{
		sb_weapons[2+i][0] = Draw_PicFromWad (va("inva%i_shotgun",i+1));
		sb_weapons[2+i][1] = Draw_PicFromWad (va("inva%i_sshotgun",i+1));
		sb_weapons[2+i][2] = Draw_PicFromWad (va("inva%i_nailgun",i+1));
		sb_weapons[2+i][3] = Draw_PicFromWad (va("inva%i_snailgun",i+1));
		sb_weapons[2+i][4] = Draw_PicFromWad (va("inva%i_rlaunch",i+1));
		sb_weapons[2+i][5] = Draw_PicFromWad (va("inva%i_srlaunch",i+1));
		sb_weapons[2+i][6] = Draw_PicFromWad (va("inva%i_lightng",i+1));
	}

	sb_ammo[0] = Draw_PicFromWad ("sb_shells");
	sb_ammo[1] = Draw_PicFromWad ("sb_nails");
	sb_ammo[2] = Draw_PicFromWad ("sb_rocket");
	sb_ammo[3] = Draw_PicFromWad ("sb_cells");

	sb_armor[0] = Draw_PicFromWad ("sb_armor1");
	sb_armor[1] = Draw_PicFromWad ("sb_armor2");
	sb_armor[2] = Draw_PicFromWad ("sb_armor3");

	sb_items[0] = Draw_PicFromWad ("sb_key1");
	sb_items[1] = Draw_PicFromWad ("sb_key2");
	sb_items[2] = Draw_PicFromWad ("sb_invis");
	sb_items[3] = Draw_PicFromWad ("sb_invuln");
	sb_items[4] = Draw_PicFromWad ("sb_suit");
	sb_items[5] = Draw_PicFromWad ("sb_quad");

	sb_sigil[0] = Draw_PicFromWad ("sb_sigil1");
	sb_sigil[1] = Draw_PicFromWad ("sb_sigil2");
	sb_sigil[2] = Draw_PicFromWad ("sb_sigil3");
	sb_sigil[3] = Draw_PicFromWad ("sb_sigil4");

	sb_faces[4][0] = Draw_PicFromWad ("face1");
	sb_faces[4][1] = Draw_PicFromWad ("face_p1");
	sb_faces[3][0] = Draw_PicFromWad ("face2");
	sb_faces[3][1] = Draw_PicFromWad ("face_p2");
	sb_faces[2][0] = Draw_PicFromWad ("face3");
	sb_faces[2][1] = Draw_PicFromWad ("face_p3");
	sb_faces[1][0] = Draw_PicFromWad ("face4");
	sb_faces[1][1] = Draw_PicFromWad ("face_p4");
	sb_faces[0][0] = Draw_PicFromWad ("face5");
	sb_faces[0][1] = Draw_PicFromWad ("face_p5");

	sb_face_invis = Draw_PicFromWad ("face_invis");
	sb_face_invuln = Draw_PicFromWad ("face_invul2");
	sb_face_invis_invuln = Draw_PicFromWad ("face_inv2");
	sb_face_quad = Draw_PicFromWad ("face_quad");

	Cmd_AddCommand ("+showscores", Sbar_ShowScores);
	Cmd_AddCommand ("-showscores", Sbar_DontShowScores);

	sb_sbar = Draw_PicFromWad ("sbar");
	sb_ibar = Draw_PicFromWad ("ibar");
	sb_scorebar = Draw_PicFromWad ("scorebar");

//MED 01/04/97 added new hipnotic weapons
	if (hipnotic)
	{
	  hsb_weapons[0][0] = Draw_PicFromWad ("inv_laser");
	  hsb_weapons[0][1] = Draw_PicFromWad ("inv_mjolnir");
	  hsb_weapons[0][2] = Draw_PicFromWad ("inv_gren_prox");
	  hsb_weapons[0][3] = Draw_PicFromWad ("inv_prox_gren");
	  hsb_weapons[0][4] = Draw_PicFromWad ("inv_prox");

	  hsb_weapons[1][0] = Draw_PicFromWad ("inv2_laser");
	  hsb_weapons[1][1] = Draw_PicFromWad ("inv2_mjolnir");
	  hsb_weapons[1][2] = Draw_PicFromWad ("inv2_gren_prox");
	  hsb_weapons[1][3] = Draw_PicFromWad ("inv2_prox_gren");
	  hsb_weapons[1][4] = Draw_PicFromWad ("inv2_prox");

	  for (i=0 ; i<5 ; i++)
	  {
		 hsb_weapons[2+i][0] = Draw_PicFromWad (va("inva%i_laser",i+1));
		 hsb_weapons[2+i][1] = Draw_PicFromWad (va("inva%i_mjolnir",i+1));
		 hsb_weapons[2+i][2] = Draw_PicFromWad (va("inva%i_gren_prox",i+1));
		 hsb_weapons[2+i][3] = Draw_PicFromWad (va("inva%i_prox_gren",i+1));
		 hsb_weapons[2+i][4] = Draw_PicFromWad (va("inva%i_prox",i+1));
	  }

	  hsb_items[0] = Draw_PicFromWad ("sb_wsuit");
	  hsb_items[1] = Draw_PicFromWad ("sb_eshld");
	}

	if (rogue)
	{
		rsb_invbar[0] = Draw_PicFromWad ("r_invbar1");
		rsb_invbar[1] = Draw_PicFromWad ("r_invbar2");

		rsb_weapons[0] = Draw_PicFromWad ("r_lava");
		rsb_weapons[1] = Draw_PicFromWad ("r_superlava");
		rsb_weapons[2] = Draw_PicFromWad ("r_gren");
		rsb_weapons[3] = Draw_PicFromWad ("r_multirock");
		rsb_weapons[4] = Draw_PicFromWad ("r_plasma");

		rsb_items[0] = Draw_PicFromWad ("r_shield1");
        rsb_items[1] = Draw_PicFromWad ("r_agrav1");

// PGM 01/19/97 - team color border
        rsb_teambord = Draw_PicFromWad ("r_teambord");
// PGM 01/19/97 - team color border

		rsb_ammo[0] = Draw_PicFromWad ("r_ammolava");
		rsb_ammo[1] = Draw_PicFromWad ("r_ammomulti");
		rsb_ammo[2] = Draw_PicFromWad ("r_ammoplasma");
	}

	R_InitXHairTexture ();
}


/*
=============
Sbar_itoa
=============
*/
int Sbar_itoa (int num, char *buf)
{
	char	*str;
	int		pow10;
	int		dig;

	str = buf;

	if (num < 0)
	{
		*str++ = '-';
		num = -num;
	}

	for (pow10 = 10 ; num >= pow10 ; pow10 *= 10)
	;

	do
	{
		pow10 /= 10;
		dig = num/pow10;
		*str++ = '0'+dig;
		num -= dig*pow10;
	} while (pow10 != 1);

	*str = 0;

	return str-buf;
}


//=============================================================================

// drawing routines are relative to the status bar location

/*
=============
Sbar_DrawPic
=============
*/
void Sbar_DrawPic (int x, int y, qpic_t *pic)
{
	if (cl.gametype == GAME_DEATHMATCH)
		Draw_Pic (x /* + ((vid.width - 320)>>1)*/, y + (vid.height-SBAR_HEIGHT), pic);
	else
		Draw_Pic (x + ((vid.width - 320)>>1), y + (vid.height-SBAR_HEIGHT), pic);
}

/*
=============
Sbar_DrawTransPic
=============
*/
void Sbar_DrawTransPic (int x, int y, qpic_t *pic)
{
	if (cl.gametype == GAME_DEATHMATCH)
		Draw_TransPic (x /*+ ((vid.width - 320)>>1)*/, y + (vid.height-SBAR_HEIGHT), pic);
	else
		Draw_TransPic (x + ((vid.width - 320)>>1), y + (vid.height-SBAR_HEIGHT), pic);
}

/*
================
Sbar_DrawCharacter

Draws one solid graphics character
================
*/
void Sbar_DrawCharacter (int x, int y, int num)
{
	if (cl.gametype == GAME_DEATHMATCH)
		Draw_Character ( x /*+ ((vid.width - 320)>>1) */ + 4 , y + vid.height-SBAR_HEIGHT, num);
	else
		Draw_Character ( x + ((vid.width - 320)>>1) + 4 , y + vid.height-SBAR_HEIGHT, num);
}

/*
================
Sbar_DrawString
================
*/
void Sbar_DrawString (int x, int y, char *str)
{
	if (cl.gametype == GAME_DEATHMATCH)
		Draw_String (x /*+ ((vid.width - 320)>>1)*/, y+ vid.height-SBAR_HEIGHT, str);
	else
		Draw_String (x + ((vid.width - 320)>>1), y+ vid.height-SBAR_HEIGHT, str);
}

/*
=============
Sbar_DrawNum
=============
*/
void Sbar_DrawNum (int x, int y, int num, int digits, int color)
{
	char			str[12];
	char			*ptr;
	int				l, frame;

	l = Sbar_itoa (num, str);
	ptr = str;
	if (l > digits)
		ptr += (l-digits);
	if (l < digits)
		x += (digits-l)*24;

	while (*ptr)
	{
		if (*ptr == '-')
			frame = STAT_MINUS;
		else
			frame = *ptr -'0';

		Sbar_DrawTransPic (x,y,sb_nums[color][frame]);
		x += 24;
		ptr++;
	}
}

void Draw_ZoomPic (int x, int y, qpic_t *pic, float Zoom);

void HUD_DrawNum (int x, int y, int num, int digits, int color, float Zoom)
{
	char			str[12];
	char			*ptr;
	int				l, frame;

	l = Sbar_itoa (num, str);

	ptr = str;

	if (l > digits) ptr += (l-digits);
	if (l < digits) x += (digits-l) * (24 * Zoom);

	if (num < -99)
		num = -99;
	if (num < 0)
		;
	else if (num < 10) x -= 24;
	else if (num < 100) x -= 12;

	while (*ptr)
	{
		if (*ptr == '-')
			frame = STAT_MINUS;
		else
			frame = *ptr -'0';

		Draw_ZoomPic (x, y, sb_nums[color][frame], Zoom);

		x += (24 * Zoom);
		ptr++;
	}
}


//=============================================================================

int		fragsort[MAX_SCOREBOARD];

char	scoreboardtext[MAX_SCOREBOARD][20];
int		scoreboardtop[MAX_SCOREBOARD];
int		scoreboardbottom[MAX_SCOREBOARD];
int		scoreboardcount[MAX_SCOREBOARD];
int		scoreboardlines;

/*
===============
Sbar_SortFrags
===============
*/
void Sbar_SortFrags (void)
{
	int		i, j, k;

// sort by frags
	scoreboardlines = 0;
	for (i=0 ; i<cl.maxclients ; i++)
	{
		if (cl.scores[i].name[0])
		{
			fragsort[scoreboardlines] = i;
			scoreboardlines++;
		}
	}

	for (i=0 ; i<scoreboardlines ; i++)
		for (j=0 ; j<scoreboardlines-1-i ; j++)
			if (cl.scores[fragsort[j]].frags < cl.scores[fragsort[j+1]].frags)
			{
				k = fragsort[j];
				fragsort[j] = fragsort[j+1];
				fragsort[j+1] = k;
			}
}

int	Sbar_ColorForMap (int m)
{
	return m < 128 ? m + 8 : m + 8;
}

/*
===============
Sbar_UpdateScoreboard
===============
*/
void Sbar_UpdateScoreboard (void)
{
	int		i, k;
	int		top, bottom;
	scoreboard_t	*s;

	Sbar_SortFrags ();

// draw the text
	memset (scoreboardtext, 0, sizeof(scoreboardtext));

	for (i=0 ; i<scoreboardlines; i++)
	{
		k = fragsort[i];
		s = &cl.scores[k];
		sprintf (&scoreboardtext[i][1], "%3i %s", s->frags, s->name);

		top = s->colors & 0xf0;
		bottom = (s->colors & 15) <<4;
		scoreboardtop[i] = Sbar_ColorForMap (top);
		scoreboardbottom[i] = Sbar_ColorForMap (bottom);
	}
}



/*
===============
Sbar_SoloScoreboard
===============
*/
void Sbar_SoloScoreboard (void)
{
	char	str[80];
	int		minutes, seconds, tens, units;
	int		l;

	sprintf (str,"Monsters:%3i /%3i", cl.stats[STAT_MONSTERS], cl.stats[STAT_TOTALMONSTERS]);
	Sbar_DrawString (8, 4, str);

	sprintf (str,"Secrets :%3i /%3i", cl.stats[STAT_SECRETS], cl.stats[STAT_TOTALSECRETS]);
	Sbar_DrawString (8, 12, str);

// time
	minutes = cl.time / 60;
	seconds = cl.time - 60*minutes;
	tens = seconds / 10;
	units = seconds - 10*tens;
	sprintf (str,"Time :%3i:%i%i", minutes, tens, units);
	Sbar_DrawString (184, 4, str);

// draw level name
	l = strlen (cl.levelname);
	Sbar_DrawString (232 - l*4, 12, cl.levelname);
}

/*
===============
Sbar_DrawScoreboard
===============
*/
void Sbar_DrawScoreboard (void)
{
	Sbar_SoloScoreboard ();
	if (cl.gametype == GAME_DEATHMATCH)
		Sbar_DeathmatchOverlay ();
#if 0
	int		i, j, c;
	int		x, y;
	int		l;
	int		top, bottom;
	scoreboard_t	*s;

	if (cl.gametype != GAME_DEATHMATCH)
	{
		Sbar_SoloScoreboard ();
		return;
	}

	Sbar_UpdateScoreboard ();

	l = scoreboardlines <= 6 ? scoreboardlines : 6;

	for (i=0 ; i<l ; i++)
	{
		x = 20*(i&1);
		y = i/2 * 8;

		s = &cl.scores[fragsort[i]];
		if (!s->name[0])
			continue;

	// draw background
		top = s->colors & 0xf0;
		bottom = (s->colors & 15)<<4;
		top = Sbar_ColorForMap (top);
		bottom = Sbar_ColorForMap (bottom);

		Draw_Fill ( x*8+10 + ((vid.width - 320)>>1), y + vid.height - SBAR_HEIGHT, 28, 4, top);
		Draw_Fill ( x*8+10 + ((vid.width - 320)>>1), y+4 + vid.height - SBAR_HEIGHT, 28, 4, bottom);

	// draw text
		for (j=0 ; j<20 ; j++)
		{
			c = scoreboardtext[i][j];
			if (c == 0 || c == ' ')
				continue;
			Sbar_DrawCharacter ( (x+j)*8, y, c);
		}
	}
#endif
}

//=============================================================================

/*
===============
Sbar_DrawInventory
===============
*/
void Sbar_DrawInventory (void)
{
	int		i;
	char	num[6];
	float	time;
	int		flashon;

	/*
	if (rogue)
	{
		if ( cl.stats[STAT_ACTIVEWEAPON] >= RIT_LAVA_NAILGUN )
			Sbar_DrawPic (0, -24, rsb_invbar[0]);
		else
			Sbar_DrawPic (0, -24, rsb_invbar[1]);
	}
	else
	{
		Sbar_DrawPic (0, -24, sb_ibar);
	}
	*/

// weapons
	for (i=0 ; i<7 ; i++)
	{
		if (cl.items & (IT_SHOTGUN<<i) )
		{
			time = cl.item_gettime[i];
			flashon = (int)((cl.time - time)*10);
			if (flashon >= 10)
			{
				if ( cl.stats[STAT_ACTIVEWEAPON] == (IT_SHOTGUN<<i)  )
					flashon = 1;
				else
					flashon = 0;
			}
			else
				flashon = (flashon%5) + 2;

         Sbar_DrawPic (i*24, -16, sb_weapons[flashon][i]);

			if (flashon > 1)
				sb_updates = 0;		// force update to remove flash
		}
	}

// MED 01/04/97
// hipnotic weapons
    if (hipnotic)
    {
      int grenadeflashing=0;
      for (i=0 ; i<4 ; i++)
      {
         if (cl.items & (1<<hipweapons[i]) )
         {
            time = cl.item_gettime[hipweapons[i]];
            flashon = (int)((cl.time - time)*10);
            if (flashon >= 10)
            {
               if ( cl.stats[STAT_ACTIVEWEAPON] == (1<<hipweapons[i])  )
                  flashon = 1;
               else
                  flashon = 0;
            }
            else
               flashon = (flashon%5) + 2;

            // check grenade launcher
            if (i==2)
            {
               if (cl.items & HIT_PROXIMITY_GUN)
               {
                  if (flashon)
                  {
                     grenadeflashing = 1;
                     Sbar_DrawPic (96, -16, hsb_weapons[flashon][2]);
                  }
               }
            }
            else if (i==3)
            {
               if (cl.items & (IT_SHOTGUN<<4))
               {
                  if (flashon && !grenadeflashing)
                  {
                     Sbar_DrawPic (96, -16, hsb_weapons[flashon][3]);
                  }
                  else if (!grenadeflashing)
                  {
                     Sbar_DrawPic (96, -16, hsb_weapons[0][3]);
                  }
               }
               else
                  Sbar_DrawPic (96, -16, hsb_weapons[flashon][4]);
            }
            else
               Sbar_DrawPic (176 + (i*24), -16, hsb_weapons[flashon][i]);
            if (flashon > 1)
               sb_updates = 0;      // force update to remove flash
         }
      }
    }

	if (rogue)
	{
    // check for powered up weapon.
		if ( cl.stats[STAT_ACTIVEWEAPON] >= RIT_LAVA_NAILGUN )
		{
			for (i=0;i<5;i++)
			{
				if (cl.stats[STAT_ACTIVEWEAPON] == (RIT_LAVA_NAILGUN << i))
				{
					Sbar_DrawPic ((i+2)*24, -16, rsb_weapons[i]);
				}
			}
		}
	}

// ammo counts
	for (i=0 ; i<4 ; i++)
	{
		sprintf (num, "%3i",cl.stats[STAT_SHELLS+i] );
		if (num[0] != ' ')
			Sbar_DrawCharacter ( (6*i+1)*8 - 2, -24, 18 + num[0] - '0');
		if (num[1] != ' ')
			Sbar_DrawCharacter ( (6*i+2)*8 - 2, -24, 18 + num[1] - '0');
		if (num[2] != ' ')
			Sbar_DrawCharacter ( (6*i+3)*8 - 2, -24, 18 + num[2] - '0');
	}

	flashon = 0;
   // items
   for (i=0 ; i<6 ; i++)
      if (cl.items & (1<<(17+i)))
      {
         time = cl.item_gettime[17+i];
         if (time && time > cl.time - 2 && flashon )
         {  // flash frame
            sb_updates = 0;
         }
         else
         {
         //MED 01/04/97 changed keys
            if (!hipnotic || (i>1))
            {
               Sbar_DrawPic (192 + i*16, -16, sb_items[i]);
            }
         }
         if (time && time > cl.time - 2)
            sb_updates = 0;
      }
   //MED 01/04/97 added hipnotic items
   // hipnotic items
   if (hipnotic)
   {
      for (i=0 ; i<2 ; i++)
         if (cl.items & (1<<(24+i)))
         {
            time = cl.item_gettime[24+i];
            if (time && time > cl.time - 2 && flashon )
            {  // flash frame
               sb_updates = 0;
            }
            else
            {
               Sbar_DrawPic (288 + i*16, -16, hsb_items[i]);
            }
            if (time && time > cl.time - 2)
               sb_updates = 0;
         }
   }

	if (rogue)
	{
	// new rogue items
		for (i=0 ; i<2 ; i++)
		{
			if (cl.items & (1<<(29+i)))
			{
				time = cl.item_gettime[29+i];

				if (time &&	time > cl.time - 2 && flashon )
				{	// flash frame
					sb_updates = 0;
				}
				else
				{
					Sbar_DrawPic (288 + i*16, -16, rsb_items[i]);
				}

				if (time &&	time > cl.time - 2)
					sb_updates = 0;
			}
		}
	}
	else
	{
	// sigils
		for (i=0 ; i<4 ; i++)
		{
			if (cl.items & (1<<(28+i)))
			{
				time = cl.item_gettime[28+i];
				if (time &&	time > cl.time - 2 && flashon )
				{	// flash frame
					sb_updates = 0;
				}
				else
					Sbar_DrawPic (320-32 + i*8, -16, sb_sigil[i]);
				if (time &&	time > cl.time - 2)
					sb_updates = 0;
			}
		}
	}
}

//=============================================================================

/*
===============
Sbar_DrawFrags
===============
*/
void Sbar_DrawFrags (void)
{
	int				i, k, l;
	int				top, bottom;
	int				x, y, f;
	int				xofs;
	char			num[12];
	scoreboard_t	*s;

	Sbar_SortFrags ();

// draw the text
	l = scoreboardlines <= 4 ? scoreboardlines : 4;

	x = 23;
	if (cl.gametype == GAME_DEATHMATCH)
		xofs = 0;
	else
		xofs = (vid.width - 320)>>1;
	y = vid.height - SBAR_HEIGHT - 23;

	for (i=0 ; i<l ; i++)
	{
		k = fragsort[i];
		s = &cl.scores[k];
		if (!s->name[0])
			continue;

	// draw background
		top = s->colors & 0xf0;
		bottom = (s->colors & 15)<<4;
		top = Sbar_ColorForMap (top);
		bottom = Sbar_ColorForMap (bottom);

		Draw_Fill (xofs + x*8 + 10, y, 28, 4, top);
		Draw_Fill (xofs + x*8 + 10, y+4, 28, 3, bottom);

	// draw number
		f = s->frags;
		sprintf (num, "%3i",f);

		Sbar_DrawCharacter ( (x+1)*8 , -24, num[0]);
		Sbar_DrawCharacter ( (x+2)*8 , -24, num[1]);
		Sbar_DrawCharacter ( (x+3)*8 , -24, num[2]);

		if (k == cl.viewentity - 1)
		{
			Sbar_DrawCharacter (x*8+2, -24, 16);
			Sbar_DrawCharacter ( (x+4)*8-4, -24, 17);
		}
		x+=4;
	}
}

//=============================================================================


/*
===============
Sbar_DrawFace
===============
*/
void Sbar_DrawFace (void)
{
	int		f, anim;

	if (r_hud.value) return;

	if ((cl.items & (IT_INVISIBILITY|IT_INVULNERABILITY)) == (IT_INVISIBILITY|IT_INVULNERABILITY))
	{
		Draw_Pic (6, vid.height - 30, sb_face_invis_invuln);
		return;
	}
	if (cl.items & IT_QUAD)
	{
		Draw_Pic (6, vid.height - 30, sb_face_quad);
		return;
	}
	if (cl.items & IT_INVISIBILITY)
	{
		Draw_Pic (6, vid.height - 30, sb_face_invis);
		return;
	}
	if (cl.items & IT_INVULNERABILITY)
	{
		Draw_Pic (6, vid.height - 30, sb_face_invuln);
		return;
	}

	// MHQuake bug
	// if health was less than -20 or so we would have a crash
	// (dvision by 20 below would have given us a negative array index - OUCH!!!)
	if (cl.stats[STAT_HEALTH] >= 100)
		f = 4;
	else if (cl.stats[STAT_HEALTH] <= 0)
		f = 0;
	else
		f = cl.stats[STAT_HEALTH] / 20;

	if (cl.time <= cl.faceanimtime)
	{
		anim = 1;
		sb_updates = 0;		// make sure the anim gets drawn over
	}
	else
		anim = 0;

	Draw_Pic (6, vid.height - 30, sb_faces[f][anim]);
}


void Sbar_DrawArmor (void)
{
	qpic_t *armor_pic;
	int Origin[2];
	float Zoom;

	if (r_hud.value)
	{
		Origin[0] = 120;
		Origin[1] = vid.height - 30;
		Zoom = 1.0;
	}
	else
	{
		Origin[0] = 36;
		Origin[1] = vid.height - 60;
		Zoom = 1.0;
	}

	if (cl.items & IT_INVULNERABILITY)
	{
		HUD_DrawNum (Origin[0], 
					 Origin[1], 
					 666, 
					 3, 
					 1,
					 Zoom);

		armor_pic = draw_disc;
	}
	else if (!cl.stats[STAT_ARMOR])
		return;
	else
	{
		HUD_DrawNum (Origin[0], 
					 Origin[1], 
					 cl.stats[STAT_ARMOR], 
					 3, 
					 cl.stats[STAT_ARMOR] <= 25,
					 Zoom);

		if (rogue)
		{
			if (cl.items & RIT_ARMOR3)
				armor_pic = sb_armor[2];
			else if (cl.items & RIT_ARMOR2)
				armor_pic = sb_armor[1];
			else if (cl.items & RIT_ARMOR1)
				armor_pic = sb_armor[0];
		}
		else
		{
			if (cl.items & IT_ARMOR3)
				armor_pic = sb_armor[2];
			else if (cl.items & IT_ARMOR2)
				armor_pic = sb_armor[1];
			else if (cl.items & IT_ARMOR1)
				armor_pic = sb_armor[0];
		}
	}

	if (r_hud.value) return;

	Draw_Pic (6, vid.height - 60, armor_pic);
}


void Sbar_DrawAmmo (void)
{
	qpic_t *ammo_pic = NULL;
	int ammo_pos;

	// ammo icon
	if (rogue)
	{
		if (cl.items & RIT_SHELLS)
			ammo_pic = sb_ammo[0];
		else if (cl.items & RIT_NAILS)
			ammo_pic = sb_ammo[1];
		else if (cl.items & RIT_ROCKETS)
			ammo_pic = sb_ammo[2];
		else if (cl.items & RIT_CELLS)
			ammo_pic = sb_ammo[3];
		else if (cl.items & RIT_LAVA_NAILS)
			ammo_pic = rsb_ammo[0];
		else if (cl.items & RIT_PLASMA_AMMO)
			ammo_pic = rsb_ammo[1];
		else if (cl.items & RIT_MULTI_ROCKETS)
			ammo_pic = rsb_ammo[2];
	}
	else
	{
		if (cl.items & IT_SHELLS)
			ammo_pic = sb_ammo[0];
		else if (cl.items & IT_NAILS)
			ammo_pic = sb_ammo[1];
		else if (cl.items & IT_ROCKETS)
			ammo_pic = sb_ammo[2];
		else if (cl.items & IT_CELLS)
			ammo_pic = sb_ammo[3];
	}

	if (!ammo_pic) return;

	if (r_hud.value)
		ammo_pos = vid.width - 78;
	else
	{
		ammo_pos = vid.width - 108;
		Draw_Pic (vid.width - 30, vid.height - 30, ammo_pic);
	}

	HUD_DrawNum (ammo_pos, vid.height - 30, cl.stats[STAT_AMMO], 3, cl.stats[STAT_AMMO] <= 10, 1.0);
	Draw_XHair (xhair_xpos, xhair_ypos);
}


void HUD_DrawInventory (void)
{
	int flashon;
	int i;
	char num[6];
	int ammo_count;

	if (r_hud.value) return;

	// weaponses
	if (rogue)
	{
		for (i=0 ; i<7 ; i++)
		{
			// this is horrible stuff.
			if ((i == 2) && (cl.stats[STAT_ACTIVEWEAPON] == RIT_LAVA_NAILGUN) && (cl.items & RIT_LAVA_NAILGUN))
		        Draw_Pic (30, 46, rsb_weapons[0]);
			else if ((i == 3) && (cl.stats[STAT_ACTIVEWEAPON] == RIT_LAVA_SUPER_NAILGUN) && (cl.items & RIT_LAVA_SUPER_NAILGUN))
		        Draw_Pic (30, 66, rsb_weapons[1]);
			else if ((i == 4) && (cl.stats[STAT_ACTIVEWEAPON] == RIT_MULTI_GRENADE) && (cl.items & RIT_MULTI_GRENADE))
		        Draw_Pic (30, 86, rsb_weapons[2]);
			else if ((i == 5) && (cl.stats[STAT_ACTIVEWEAPON] == RIT_MULTI_ROCKET) && (cl.items & RIT_MULTI_ROCKET))
		        Draw_Pic (30, 106, rsb_weapons[3]);
			else if ((i == 6) && (cl.stats[STAT_ACTIVEWEAPON] == RIT_PLASMA_GUN) && (cl.items & RIT_PLASMA_GUN))
		        Draw_Pic (6, 126, rsb_weapons[4]);
			else if (cl.items & (IT_SHOTGUN<<i) )
			{
				flashon = (cl.stats[STAT_ACTIVEWEAPON] == (IT_SHOTGUN<<i));
		        if (i == 6) Draw_Pic (6, 6+(i*20), sb_weapons[flashon][i]);
				else Draw_Pic (30, 6+(i*20), sb_weapons[flashon][i]);
			}
			else
				continue;

			ammo_count = i/2;  // because it's an int, it will truncate instead of round up
			sprintf (num, "%3i",cl.stats[STAT_SHELLS+ammo_count]);

			if (num[0] != ' ') Draw_Character ( 60, 8+(i*20), 176 + num[0] - '0');
			if (num[1] != ' ') Draw_Character ( 68, 8+(i*20), 176 + num[1] - '0');
			if (num[2] != ' ') Draw_Character ( 76, 8+(i*20), 176 + num[2] - '0');
		}
	}
	else
	{
		for (i=0 ; i<7 ; i++)
		{
			if (cl.items & (IT_SHOTGUN<<i) )
			{
				flashon = (cl.stats[STAT_ACTIVEWEAPON] == (IT_SHOTGUN<<i));
		        if (i == 6 && !hipnotic) Draw_Pic (6, 6+(i*20), sb_weapons[flashon][i]);
				else if (i == 6 && hipnotic) Draw_Pic (22, 6+(i*20), sb_weapons[flashon][i]);
				else Draw_Pic (30, 6+(i*20), sb_weapons[flashon][i]);

				// ammo counts
				ammo_count = i/2;  // because it's an int, it will truncate instead of round up
				sprintf (num, "%3i",cl.stats[STAT_SHELLS+ammo_count]);

				if (num[0] != ' ') Draw_Character ( 60, 8+(i*20), 176 + num[0] - '0');
				if (num[1] != ' ') Draw_Character ( 68, 8+(i*20), 176 + num[1] - '0');
				if (num[2] != ' ') Draw_Character ( 76, 8+(i*20), 176 + num[2] - '0');
			}
		}
	}

	// hipnotic extra weapons
	if (hipnotic)
	{
		// proxy gun - alt for weapon 6 - this is horrible
		if (cl.items & HIT_PROXIMITY_GUN)
		{
			if (cl.stats[STAT_ACTIVEWEAPON] == HIT_PROXIMITY_GUN)
			{
		        Draw_Pic (30, 86, hsb_weapons[1][2]);
				sprintf (num, "%3i",cl.stats[STAT_ROCKETS]);
				if (num[0] != ' ') Draw_Character ( 60, 88, 176 + num[0] - '0');
				if (num[1] != ' ') Draw_Character ( 68, 88, 176 + num[1] - '0');
				if (num[2] != ' ') Draw_Character ( 76, 88, 176 + num[2] - '0');
			}
			else if (!(cl.items & IT_GRENADE_LAUNCHER))
			{
		        Draw_Pic (30, 86, hsb_weapons[0][2]);
				sprintf (num, "%3i",cl.stats[STAT_ROCKETS]);
				if (num[0] != ' ') Draw_Character ( 60, 88, 176 + num[0] - '0');
				if (num[1] != ' ') Draw_Character ( 68, 88, 176 + num[1] - '0');
				if (num[2] != ' ') Draw_Character ( 76, 88, 176 + num[2] - '0');
			}
		}

		// mjolnir - weapon 0
		if (cl.items & HIT_MJOLNIR)
		{
			flashon = (cl.stats[STAT_ACTIVEWEAPON] == HIT_MJOLNIR);
	        Draw_Pic (30, 166, hsb_weapons[flashon][1]);
			sprintf (num, "%3i",cl.stats[STAT_CELLS]);
			if (num[0] != ' ') Draw_Character ( 60, 168, 176 + num[0] - '0');
			if (num[1] != ' ') Draw_Character ( 68, 168, 176 + num[1] - '0');
			if (num[2] != ' ') Draw_Character ( 76, 168, 176 + num[2] - '0');
		}

		// laser cannon - weapon 9
		if (cl.items & HIT_LASER_CANNON)
		{
			flashon = (cl.stats[STAT_ACTIVEWEAPON] == HIT_LASER_CANNON);
	        Draw_Pic (30, 146, hsb_weapons[flashon][0]);
			sprintf (num, "%3i",cl.stats[STAT_CELLS]);
			if (num[0] != ' ') Draw_Character ( 60, 148, 176 + num[0] - '0');
			if (num[1] != ' ') Draw_Character ( 68, 148, 176 + num[1] - '0');
			if (num[2] != ' ') Draw_Character ( 76, 148, 176 + num[2] - '0');
		}
	}
}


void R_DrawAliasModelSBar (entity_t *e, float Light, int skin, float trans);
void M_DrawTextBoxFreeAlpha (int x, int y, int width, int lines, float alpha);

// no g_ for the standard shotgun
char *Weaponses[] =
{
	// normal quake
	"progs/g_shot.mdl",
	"progs/g_shot.mdl",
	"progs/g_nail.mdl",
	"progs/g_nail2.mdl",
	"progs/g_rock.mdl",
	"progs/g_rock2.mdl",
	"progs/g_light.mdl",
	// hipnotic
	"progs/g_hammer.mdl",
	"progs/g_laserg.mdl",
	"progs/g_prox.mdl"
	// no g_*.mdl files for rogue
};


/*
=========================
SBar_SetViewportFor3DModel

Generates a new viewport of width 200 and height relative to the screen's aspect ratio
for drawing a 3D model into.  The model will be centered in the origin of the viewport
meaning that we won't get any perspective distortion.  The co-ordinates x and y are 
relative to the bottom-left of the screen (you'll have to calculate these yourself).

Also initializes new matrices and puts Z going up, so it's the same as Quake.

NOTE: Some OpenGL implementations DO NOT support viewports which are partially outside
of the main screen are.  If this happens to you (and you'll know it because the HUD
items won't be properly lined up) it's because your OpenGL implementation is CRAP, and
NOT because of an MHQuake bug.  Suggested solutions are:

* update your drivers.
* stop using beta drivers.
* stop using 3DFX cards.
* file a bug report with your 3D card manufacturer.

I will ONLY work to resolve bugs in MHQuake - resolution of crap 3D card issues is
NOT MY PROBLEM.
=========================
*/
void MYgluPerspective (GLdouble fovy, GLdouble aspect, GLdouble zNear, GLdouble zFar);

void Sbar_SetViewportFor3DModel (int x, int y)
{
	int vpWidth = 200;
	int vpHeight;

	vpHeight = (vpWidth * glheight) / glwidth;

	glViewport (x, y, vpWidth, vpHeight);
	//glViewport (glx, gly, glwidth, glheight);

	// initialize our matrices
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();

	// we don't really need a huge far clipping plane here
	MYgluPerspective (r_refdef.fov_y, (float) vpWidth / (float) vpHeight, 4, 150);

	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();

	// we don't want OpenGL to remove these if they're deeper than pixels in the world
	glDisable (GL_DEPTH_TEST);

	// put Z going up and X going deep to match Quakes main render
	// Y remains at right to left, Z becomes top to bottom and X becomes near to far
	glRotatef (-90, 1, 0, 0);
	glRotatef (90, 0, 0, 1);
}


model_t *InvWeaps[10] = {NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL};

void DrawInvWeap (int WeapNum, float Colour, qboolean active)
{
	entity_t e;
	float bobjrotate = anglemod (100 * cl.time);

	// generate our viewport
	Sbar_SetViewportFor3DModel (-50, glheight - 100 - (25 * WeapNum));

	// set up a temp dummy entity to hold the weapon
	memset (&e, 0, sizeof (e));
	e.model = InvWeaps[WeapNum];

	// set origin and angles for any transforms we might want to do
	e.origin[0] = 85;		// always positive, higher numbers are deeper
	e.origin[1] = 0;		// positive = to the left, negative = to the right
	e.origin[2] = -36;		// negative = further down, positive = higher up
	e.angles[1] = ((WeapNum & 1) * 45) + bobjrotate;

	if (active)
		e.origin[0] -= 5;
	else e.origin[0] += 15;

	// and draw it
	R_DrawAliasModelSBar (&e, Colour, 0, 1.0);
}


void HUD_DrawWeapons (void)
{
	int i;

	// because rogue has no g_ weapons, we just stick with the classics for that
	// weapons are drawn in 3D but ammo counts are drawn in 2D, so they have to be done
	// seperately...
	for (i = 0; i < 7; i++)
	{
		if (cl.items & (IT_SHOTGUN << i))
		{
			if (!InvWeaps[i]) continue;
	        if (i == 6 && hipnotic) continue;

			DrawInvWeap (i, 0.7, (cl.stats[STAT_ACTIVEWEAPON] == (IT_SHOTGUN << i)));
		}
	}

	// hipnotic extra weapons
	if (hipnotic)
	{
		// snowball's chance...
	}
}


void HUD_DrawAmmoCount (void)
{
	int flashon;
	int i;
	char num[6];
	int ammo_count;

	// weaponses
	if (rogue)
	{
		// forget it, mate
	}
	else
	{
		for (i=0 ; i<7 ; i++)
		{
			if (cl.items & (IT_SHOTGUN << i) )
			{
				ammo_count = i / 2;  // because it's an int, it will truncate instead of round up
				sprintf (num, "%3i", cl.stats[STAT_SHELLS + ammo_count]);

				if (num[0] != ' ') Draw_Character (54, 27 + (i * 26), 176 + num[0] - '0');
				if (num[1] != ' ') Draw_Character (62, 27 + (i * 26), 176 + num[1] - '0');
				if (num[2] != ' ') Draw_Character (70, 27 + (i * 26), 176 + num[2] - '0');
			}
		}
	}

	if (hipnotic)
	{
		// not a hope in hell...
	}
}


model_t *HUD_Arm = NULL;

void HUD_DrawArmour (void)
{
	int skinnum;
	model_t *Armour;

	entity_t e;
	float bobjrotate = anglemod (100 * cl.time);

	// HUD replacement for Sbar_DrawArmor
	if (!cl.stats[STAT_ARMOR])
		return;
	else
	{
		if (!HUD_Arm) return;

		if (cl.items & IT_ARMOR3)
			skinnum = 2;
		else if (cl.items & IT_ARMOR2)
			skinnum = 1;
		else if (cl.items & IT_ARMOR1)
			skinnum = 0;

		Armour = HUD_Arm;
	}

	Sbar_SetViewportFor3DModel (60, 4);

	// set up a temp dummy entity to hold the model
	memset (&e, 0, sizeof (e));
	e.model = Armour;

	// set origin and angles for any transforms we might want to do
	e.origin[0] = 80;		// always positive, higher numbers are deeper
	e.origin[1] = 0;		// positive = to the left, negative = to the right
	e.origin[2] = -36;		// negative = further down, positive = higher up
	e.angles[1] = bobjrotate;

	// and draw it
	R_DrawAliasModelSBar (&e, 0.7, skinnum, 1.0);
}


char *HUD_AmmoNames[] =
{
	// normal quake
	"maps/b_shell1.bsp",
	"maps/b_nail1.bsp",
	"maps/b_rock1.bsp",
	"maps/b_batt1.bsp",
	// rogue
	"maps/b_lnail1.bsp",
	"maps/b_mrock1.bsp",
	"maps/b_plas1.bsp"
	// hipnotic has no ammo box models
};


model_t *HUD_Ammo[7] = {NULL, NULL, NULL, NULL, NULL, NULL, NULL};

void R_DrawBrushModelSbar (entity_t *e, float Light);
float swayLR = 0;
int swayLRDir = 1;
extern float frametime;

void HUD_DrawAmmo (void)
{
	entity_t e;
	int AmmoNum;

	// HUD Replacement for Sbar_DrawAmmo
	if (cl.items & IT_SHELLS)
		AmmoNum = 0;
	else if (cl.items & IT_NAILS)
		AmmoNum = 1;
	else if (cl.items & IT_ROCKETS)
		AmmoNum = 2;
	else if (cl.items & IT_CELLS)
		AmmoNum = 3;
	else return;

	if (!HUD_Ammo[AmmoNum]) return;

	Sbar_SetViewportFor3DModel (vid.width - 201, 28);

	// set up a temp dummy entity to hold the model
	memset (&e, 0, sizeof (e));
	e.model = HUD_Ammo[AmmoNum];

	// set origin and angles for any transforms we might want to do
	// different origin effects here...
	e.origin[0] = -60;		// positive: left, negative: right
	e.origin[1] = -130;
	e.origin[2] = -55;

	e.angles[1] = 35 + swayLR;

	glRotatef (90, 0, 0, 1);

	// and draw it
	R_DrawBrushModelSbar (&e, 0.7);
}


void HUD_DrawFace (void)
{
	entity_t e;

	// HUD Replacement for Sbar_DrawFace - uses standard models already loaded
	Sbar_SetViewportFor3DModel (0, 36);

	// set up a temp dummy entity to hold the model
	memset (&e, 0, sizeof (e));

	// set origin and angles for any transforms we might want to do
	e.origin[0] = 80;		// always positive, higher numbers are deeper
	e.origin[1] = 40;		// positive = to the left, negative = to the right
	e.origin[2] = -36;		// negative = further down, positive = higher up
	e.angles[1] = -100;		// face the player

	if (cl.items & IT_INVISIBILITY)
	{
		// THE EYEEEESSSS!!!!  THE EEEEYYEEEEEESSSSS!!!!!!!!!!!!!!!!!!!!
		e.origin[0] = 10;
		e.origin[1] = 5;
		e.origin[2] = -27;
	}

	e.model = cl_entities[cl.viewentity].model;
	e.frame = cl_entities[cl.viewentity].frame;

	// and draw it
	R_DrawAliasModelSBar (&e, 0.7, 0, 1.0);
}


model_t *HUD_Invl = NULL;
model_t *HUD_Quad = NULL;
int StartX;

void HUD_DrawIndividualPowerup (model_t *HUD_Powerup, int XPos)
{
	entity_t e;
	float bobjrotate = anglemod (100 * cl.time);

	// HUD Replacement for Sbar_DrawFace
	if (!HUD_Powerup) return;

	Sbar_SetViewportFor3DModel (XPos, 0);

	// set up a temp dummy entity to hold the model
	memset (&e, 0, sizeof (e));
	e.model = HUD_Powerup;

	// set origin and angles for any transforms we might want to do
	e.origin[0] = 80;		// always positive, higher numbers are deeper
	e.origin[1] = 0;		// positive = to the left, negative = to the right
	e.origin[2] = -36;		// negative = further down, positive = higher up
	e.angles[1] = bobjrotate;

	// and draw it
	R_DrawAliasModelSBar (&e, 0.7, 0, 1.0);

	StartX += 70;
}


void HUD_DrawPowerups (void)
{
	int numPowerups = 0;

	StartX = (glwidth / 2) - 100;

	if (cl.items & IT_QUAD) numPowerups++;
	if (cl.items & IT_INVULNERABILITY) numPowerups++;
	// add r & h

	if (!numPowerups) return;

	// positioning depends on the number of powerups
	if (!(numPowerups % 2))
	{
		// even number
		StartX += 35;
	}

	StartX -= (70 * (numPowerups - 1));

	// now draw them
	if (cl.items & IT_QUAD) HUD_DrawIndividualPowerup (HUD_Quad, StartX);
	if (cl.items & IT_INVULNERABILITY) HUD_DrawIndividualPowerup (HUD_Invl, StartX);
	// add r & h
}


int BitPos;

char *HUD_BitNames[] =
{
	"progs/end1.mdl",
	"progs/end2.mdl",
	"progs/end3.mdl",
	"progs/end4.mdl",
	"progs/b_s_key.mdl",
	"progs/b_g_key.mdl",
	"progs/m_s_key.mdl",
	"progs/m_g_key.mdl",
	"progs/w_s_key.mdl",
	"progs/w_g_key.mdl"
};

model_t *HUD_BitModels[10] = {NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL};

extern int worldtype;

void HUD_DrawIndividualBit (int BitNum)
{
	entity_t e;
	float bobjrotate = anglemod (100 * cl.time);

	// switch the keys
	if (BitNum >= 4)
	{
		switch (worldtype)
		{
		case 1:
			// metal
			BitNum += 2;
			break;
		case 2:
			// base
			break;
		case 0:
		default:
			// medieval
			BitNum += 4;
			break;
		}
	}

	if (!HUD_BitModels[BitNum]) return;

	Sbar_SetViewportFor3DModel (vid.width - 210 - BitPos, 4);

	// set up a temp dummy entity to hold the model
	memset (&e, 0, sizeof (e));
	e.model = HUD_BitModels[BitNum];

	e.origin[0] = 80;		// always positive, higher numbers are deeper
	e.origin[1] = 0;		// positive = to the left, negative = to the right
	e.origin[2] = -36;		// negative = further down, positive = higher up

	// give each bit a different rotation
	e.angles[1] = bobjrotate + (15 * (BitNum + 1));

	R_DrawAliasModelSBar (&e, 0.7, 0, 1.0);

	BitPos += 30;
}


void HUD_DrawBits (void)
{
	BitPos = 0;

	// drawing keys, sigils, etc
	// keys
	if (cl.items & IT_KEY1) HUD_DrawIndividualBit (4);
	if (cl.items & IT_KEY2) HUD_DrawIndividualBit (5);

	// don't draw sigils in rogue or hipnotic
	if (rogue || hipnotic) return;

	if (cl.items & IT_SIGIL1) HUD_DrawIndividualBit (0);
	if (cl.items & IT_SIGIL2) HUD_DrawIndividualBit (1);
	if (cl.items & IT_SIGIL3) HUD_DrawIndividualBit (2);
	if (cl.items & IT_SIGIL4) HUD_DrawIndividualBit (3);
}

void Sbar_Preload3DModels (void)
{
	// preload all the models to prevent disk access at run time
	// done every map load to also prevent Quake from discarding the models or textures
	int i;

	// weapons
	for (i = 0; i < 10; i++)
		InvWeaps[i] = Mod_ForName (Weaponses[i], false);

	// bits (keys/sigils)
	for (i = 0; i < 10; i++)
		HUD_BitModels[i] = Mod_ForName (HUD_BitNames[i], false);

	// ammo - loaded as required (it messes with the world lighting)
	for (i = 0; i < 7; i++)
		HUD_Ammo[i] = Mod_ForName (HUD_AmmoNames[i], false);

	// powerups
	HUD_Invl = Mod_ForName ("progs/invulner.mdl", false);
	HUD_Quad = Mod_ForName ("progs/quaddama.mdl", false);

	// armour model
	HUD_Arm = Mod_ForName ("progs/armor.mdl", false);
}


void SbarDraw3DModels (void)
{
	int i;

	if (!r_hud.value) return;

	// console is full screen
	if (scr_con_current == vid.height) return;

	// don't draw during an intermission
	if (cl.intermission) return;

	if (!cl.stats[STAT_HEALTH]) return;

	if (swayLR > 20)
	{
		swayLRDir = -1;
		swayLR = 20;
	}

	if (swayLR < 0)
	{
		swayLRDir = 1;
		swayLR = 0;
	}

	swayLR += (frametime * 10 * swayLRDir);

	// draw the models
	HUD_DrawWeapons ();
	HUD_DrawArmour ();
	HUD_DrawFace ();
	HUD_DrawPowerups ();
	HUD_DrawAmmo ();
	HUD_DrawBits ();
}


/*
===============
Sbar_Draw
===============
*/
void M_DrawTextBoxFree (int x, int y, int width, int lines);

float	fpscounter;
char	fpsss[32];
char	leafs[32];
char	surfs[32];
extern mleaf_t *r_viewleaf;

void Sbar_DrawFPS (void)
{
	static float second_count;
	static int fps;

	second_count = second_count + frametime;

	if (!r_speeds.value) return;

	// only update 10 times per frame
	if (second_count >= 0.1)
	{
		fps = 1 / frametime;
		second_count = 0;
		sprintf (fpsss, "  FPS: %4i", fps);
		sprintf (surfs, "Surfs: %4i", addSurfs);
	}

	// always do addleafs since it updates seperately
	if (addLeafs) sprintf (leafs, "Leafs: %4i", addLeafs);

	//M_DrawTextBoxFreeAlpha (vid.width - 112, 0, 12, 5, 0.5);
	Draw_String (vid.width - 100, 12, fpsss);
	Draw_String (vid.width - 100, 24, leafs);
	Draw_String (vid.width - 100, 36, surfs);

	// just in case...
	if (r_viewleaf)
	{
		Draw_String (vid.width - 100, 52, va (" Leaf: %4i", r_viewleaf->leafnum));
		Draw_String (vid.width - 100, 64, va ("Conts: %i", r_viewleaf->contents));
		Draw_String (vid.width - 100, 76, va ("  Vol: %3i", r_viewleaf->volume));
	}
}


/*
====================
SbarDraw3DModels

MH - draws 3D models for a lot of the little status bar widgets
====================
*/
void Sbar_Draw (void)
{
	if (scr_con_current == vid.height)
		return;		// console is full screen

	// don't draw during an intermission
	if (cl.intermission) return;

	// don't draw the sbar for a QC cutscene
	//if (scr_viewsize.value > 110) return;

	scr_copyeverything = 1;

	sb_updates++;

	if (sb_showscores || cl.stats[STAT_HEALTH] <= 0)
	{
		Sbar_DrawScoreboard ();
		sb_updates = 0;
	}

	if (cl.stats[STAT_HEALTH])
	{
		// armor
		Sbar_DrawArmor ();

		// face
		Sbar_DrawFace ();

		// health
		if (r_hud.value)
		{
			HUD_DrawNum (6, 
						 vid.height - 30, 
						 cl.stats[STAT_HEALTH], 
						 3, 
						 cl.stats[STAT_HEALTH] <= 25,
						 1.0);
		}
		else
		{
			HUD_DrawNum (36, 
						 vid.height - 30, 
						 cl.stats[STAT_HEALTH], 
						 3, 
						 cl.stats[STAT_HEALTH] <= 25,
						 1.0);
		}

		// ammo
		Sbar_DrawAmmo ();

		if (!r_hud.value)
		{
			// keys
			if (cl.items & IT_KEY1) Draw_Pic (vid.width - 44, vid.height - 50, sb_items[0]);
			if (cl.items & IT_KEY2) Draw_Pic (vid.width - 22, vid.height - 50, sb_items[1]);

			// sigils
			if (!rogue && !hipnotic)
			{
				// don't draw sigils in rogue or hipnotic
				if (cl.items & IT_SIGIL1) Draw_Pic (vid.width - 64, vid.height - 50, sb_sigil[0]);
				if (cl.items & IT_SIGIL2) Draw_Pic (vid.width - 78, vid.height - 50, sb_sigil[1]);
				if (cl.items & IT_SIGIL3) Draw_Pic (vid.width - 92, vid.height - 50, sb_sigil[2]);
				if (cl.items & IT_SIGIL4) Draw_Pic (vid.width - 106, vid.height - 50, sb_sigil[3]);
			}
		}

		if (r_hud.value)
			HUD_DrawAmmoCount ();
		else HUD_DrawInventory ();

		Sbar_DrawFPS ();
	}
}

//=============================================================================

/*
==================
Sbar_IntermissionNumber

==================
*/
void Sbar_IntermissionNumber (int x, int y, int num, int digits, int color)
{
	char			str[12];
	char			*ptr;
	int				l, frame;

	l = Sbar_itoa (num, str);
	ptr = str;
	if (l > digits)
		ptr += (l-digits);
	if (l < digits)
		x += (digits-l)*24;

	while (*ptr)
	{
		if (*ptr == '-')
			frame = STAT_MINUS;
		else
			frame = *ptr -'0';

		Draw_TransPic (x,y,sb_nums[color][frame]);
		x += 24;
		ptr++;
	}
}

/*
==================
Sbar_DeathmatchOverlay

==================
*/
void Sbar_DeathmatchOverlay (void)
{
	qpic_t			*pic;
	int				i, k, l;
	int				top, bottom;
	int				x, y, f;
	char			num[12];
	scoreboard_t	*s;

	scr_copyeverything = 1;
	scr_fullupdate = 0;

	pic = Draw_CachePic ("gfx/ranking.lmp");
	M_DrawPic ((320-pic->width)/2, 8, pic);

// scores
	Sbar_SortFrags ();

// draw the text
	l = scoreboardlines;

	x = 80 + ((vid.width - 320)>>1);
	y = 40;
	for (i=0 ; i<l ; i++)
	{
		k = fragsort[i];
		s = &cl.scores[k];
		if (!s->name[0])
			continue;

	// draw background
		top = s->colors & 0xf0;
		bottom = (s->colors & 15)<<4;
		top = Sbar_ColorForMap (top);
		bottom = Sbar_ColorForMap (bottom);

		Draw_Fill ( x, y, 40, 4, top);
		Draw_Fill ( x, y+4, 40, 4, bottom);

	// draw number
		f = s->frags;
		sprintf (num, "%3i",f);

		Draw_Character ( x+8 , y, num[0]);
		Draw_Character ( x+16 , y, num[1]);
		Draw_Character ( x+24 , y, num[2]);

		if (k == cl.viewentity - 1)
			Draw_Character ( x - 8, y, 12);

#if 0
{
	int				total;
	int				n, minutes, tens, units;

	// draw time
		total = cl.completed_time - s->entertime;
		minutes = (int)total/60;
		n = total - minutes*60;
		tens = n/10;
		units = n%10;

		sprintf (num, "%3i:%i%i", minutes, tens, units);

		Draw_String ( x+48 , y, num);
}
#endif

	// draw name
		Draw_String (x+64, y, s->name);

		y += 10;
	}
}

/*
==================
Sbar_DeathmatchOverlay

==================
*/
void Sbar_MiniDeathmatchOverlay (void)
{
	qpic_t			*pic;
	int				i, k, l;
	int				top, bottom;
	int				x, y, f;
	char			num[12];
	scoreboard_t	*s;
	int				numlines;

	if (vid.width < 512 || !sb_lines)
		return;

	scr_copyeverything = 1;
	scr_fullupdate = 0;

// scores
	Sbar_SortFrags ();

// draw the text
	l = scoreboardlines;
	y = vid.height - sb_lines;
	numlines = sb_lines/8;
	if (numlines < 3)
		return;

	//find us
	for (i = 0; i < scoreboardlines; i++)
		if (fragsort[i] == cl.viewentity - 1)
			break;

    if (i == scoreboardlines) // we're not there
            i = 0;
    else // figure out start
            i = i - numlines/2;

    if (i > scoreboardlines - numlines)
            i = scoreboardlines - numlines;
    if (i < 0)
            i = 0;

	x = 324;
	for (/* */; i < scoreboardlines && y < vid.height - 8 ; i++)
	{
		k = fragsort[i];
		s = &cl.scores[k];
		if (!s->name[0])
			continue;

	// draw background
		top = s->colors & 0xf0;
		bottom = (s->colors & 15)<<4;
		top = Sbar_ColorForMap (top);
		bottom = Sbar_ColorForMap (bottom);

		Draw_Fill ( x, y+1, 40, 3, top);
		Draw_Fill ( x, y+4, 40, 4, bottom);

	// draw number
		f = s->frags;
		sprintf (num, "%3i",f);

		Draw_Character ( x+8 , y, num[0]);
		Draw_Character ( x+16 , y, num[1]);
		Draw_Character ( x+24 , y, num[2]);

		if (k == cl.viewentity - 1) {
			Draw_Character ( x, y, 16);
			Draw_Character ( x + 32, y, 17);
		}

#if 0
{
	int				total;
	int				n, minutes, tens, units;

	// draw time
		total = cl.completed_time - s->entertime;
		minutes = (int)total/60;
		n = total - minutes*60;
		tens = n/10;
		units = n%10;

		sprintf (num, "%3i:%i%i", minutes, tens, units);

		Draw_String ( x+48 , y, num);
}
#endif

	// draw name
		Draw_String (x+48, y, s->name);

		y += 8;
	}
}

/*
==================
Sbar_IntermissionOverlay

==================
*/
void Sbar_IntermissionOverlay (void)
{
	qpic_t	*pic;
	int		dig;
	int		num;

	scr_copyeverything = 1;
	scr_fullupdate = 0;

	if (cl.gametype == GAME_DEATHMATCH)
	{
		Sbar_DeathmatchOverlay ();
		return;
	}

	pic = Draw_CachePic ("gfx/complete.lmp");
	Draw_Pic (64, 24, pic);

	pic = Draw_CachePic ("gfx/inter.lmp");
	Draw_TransPic (0, 56, pic);

// time
	dig = cl.completed_time/60;
	Sbar_IntermissionNumber (160, 64, dig, 3, 0);
	num = cl.completed_time - dig*60;
	Draw_TransPic (234,64,sb_colon);
	Draw_TransPic (246,64,sb_nums[0][num/10]);
	Draw_TransPic (266,64,sb_nums[0][num%10]);

	Sbar_IntermissionNumber (160, 104, cl.stats[STAT_SECRETS], 3, 0);
	Draw_TransPic (232,104,sb_slash);
	Sbar_IntermissionNumber (240, 104, cl.stats[STAT_TOTALSECRETS], 3, 0);

	Sbar_IntermissionNumber (160, 144, cl.stats[STAT_MONSTERS], 3, 0);
	Draw_TransPic (232,144,sb_slash);
	Sbar_IntermissionNumber (240, 144, cl.stats[STAT_TOTALMONSTERS], 3, 0);

}


/*
==================
Sbar_FinaleOverlay

==================
*/
void Sbar_FinaleOverlay (void)
{
	qpic_t	*pic;

	scr_copyeverything = 1;

	pic = Draw_CachePic ("gfx/finale.lmp");
	Draw_TransPic ( (vid.width-pic->width)/2, 16, pic);
}
